from __future__ import annotations

import logging
from typing import Tuple

from .exceptions import RequestFailedException, RequestRejectedException
from .inverter import Inverter
from .inverter import OperationMode
from .inverter import SensorKind as Kind
from .modbus import ILLEGAL_DATA_ADDRESS
from .model import is_2_battery, is_4_mppt, is_745_platform, is_single_phase
from .protocol import ProtocolCommand
from .sensor import *

logger = logging.getLogger(__name__)


class ET(Inverter):
    """Class representing inverter of ET/EH/BT/BH or GE's GEH families AKA platform 205 or 745"""

    # Modbus registers from offset 0x891c (35100), count 0x7d (125)
    __all_sensors: Tuple[Sensor, ...] = (
        Timestamp("timestamp", 35100, "Timestamp"),
        Voltage("vpv1", 35103, "PV1 Voltage", Kind.PV),
        Current("ipv1", 35104, "PV1 Current", Kind.PV),
        Power4("ppv1", 35105, "PV1 Power", Kind.PV),
        Voltage("vpv2", 35107, "PV2 Voltage", Kind.PV),
        Current("ipv2", 35108, "PV2 Current", Kind.PV),
        Power4("ppv2", 35109, "PV2 Power", Kind.PV),
        Voltage("vpv3", 35111, "PV3 Voltage", Kind.PV),
        Current("ipv3", 35112, "PV3 Current", Kind.PV),
        Power4("ppv3", 35113, "PV3 Power", Kind.PV),
        Voltage("vpv4", 35115, "PV4 Voltage", Kind.PV),
        Current("ipv4", 35116, "PV4 Current", Kind.PV),
        Power4("ppv4", 35117, "PV4 Power", Kind.PV),
        # ppv1 + ppv2 + ppv3 + ppv4
        Calculated("ppv",
                   lambda data:
                   max(0, read_bytes4(data, 35105, 0)) +
                   max(0, read_bytes4(data, 35109, 0)) +
                   max(0, read_bytes4(data, 35113, 0)) +
                   max(0, read_bytes4(data, 35117, 0)),
                   "PV Power", "W", Kind.PV),
        ByteH("pv4_mode", 35119, "PV4 Mode code", "", Kind.PV),
        EnumH("pv4_mode_label", 35119, PV_MODES, "PV4 Mode", Kind.PV),
        ByteL("pv3_mode", 35119, "PV3 Mode code", "", Kind.PV),
        EnumL("pv3_mode_label", 35119, PV_MODES, "PV3 Mode", Kind.PV),
        ByteH("pv2_mode", 35120, "PV2 Mode code", "", Kind.PV),
        EnumH("pv2_mode_label", 35120, PV_MODES, "PV2 Mode", Kind.PV),
        ByteL("pv1_mode", 35120, "PV1 Mode code", "", Kind.PV),
        EnumL("pv1_mode_label", 35120, PV_MODES, "PV1 Mode", Kind.PV),
        Voltage("vgrid", 35121, "On-grid L1 Voltage", Kind.AC),
        Current("igrid", 35122, "On-grid L1 Current", Kind.AC),
        Frequency("fgrid", 35123, "On-grid L1 Frequency", Kind.AC),
        # 35124 reserved
        PowerS("pgrid", 35125, "On-grid L1 Power", Kind.AC),
        Voltage("vgrid2", 35126, "On-grid L2 Voltage", Kind.AC),
        Current("igrid2", 35127, "On-grid L2 Current", Kind.AC),
        Frequency("fgrid2", 35128, "On-grid L2 Frequency", Kind.AC),
        # 35129 reserved
        PowerS("pgrid2", 35130, "On-grid L2 Power", Kind.AC),
        Voltage("vgrid3", 35131, "On-grid L3 Voltage", Kind.AC),
        Current("igrid3", 35132, "On-grid L3 Current", Kind.AC),
        Frequency("fgrid3", 35133, "On-grid L3 Frequency", Kind.AC),
        # 35134 reserved
        PowerS("pgrid3", 35135, "On-grid L3 Power", Kind.AC),
        Integer("grid_mode", 35136, "Grid Mode code", "", Kind.PV),
        Enum2("grid_mode_label", 35136, GRID_MODES, "Grid Mode", Kind.PV),
        # 35137 reserved
        PowerS("total_inverter_power", 35138, "Total Power", Kind.AC),
        # 35139 reserved
        PowerS("active_power", 35140, "Active Power", Kind.GRID),
        Calculated("grid_in_out",
                   lambda data: read_grid_mode(data, 35140),
                   "On-grid Mode code", "", Kind.GRID),
        EnumCalculated("grid_in_out_label",
                       lambda data: read_grid_mode(data, 35140), GRID_IN_OUT_MODES,
                       "On-grid Mode", Kind.GRID),
        # 35141 reserved
        Reactive("reactive_power", 35142, "Reactive Power", Kind.GRID),
        # 35143 reserved
        Apparent("apparent_power", 35144, "Apparent Power", Kind.GRID),
        Voltage("backup_v1", 35145, "Back-up L1 Voltage", Kind.UPS),
        Current("backup_i1", 35146, "Back-up L1 Current", Kind.UPS),
        Frequency("backup_f1", 35147, "Back-up L1 Frequency", Kind.UPS),
        Integer("load_mode1", 35148, "Load Mode L1"),
        # 35149 reserved
        PowerS("backup_p1", 35150, "Back-up L1 Power", Kind.UPS),
        Voltage("backup_v2", 35151, "Back-up L2 Voltage", Kind.UPS),
        Current("backup_i2", 35152, "Back-up L2 Current", Kind.UPS),
        Frequency("backup_f2", 35153, "Back-up L2 Frequency", Kind.UPS),
        Integer("load_mode2", 35154, "Load Mode L2"),
        # 35155 reserved
        PowerS("backup_p2", 35156, "Back-up L2 Power", Kind.UPS),
        Voltage("backup_v3", 35157, "Back-up L3 Voltage", Kind.UPS),
        Current("backup_i3", 35158, "Back-up L3 Current", Kind.UPS),
        Frequency("backup_f3", 35159, "Back-up L3 Frequency", Kind.UPS),
        Integer("load_mode3", 35160, "Load Mode L3"),
        # 35161 reserved
        PowerS("backup_p3", 35162, "Back-up L3 Power", Kind.UPS),
        # 35163 reserved
        PowerS("load_p1", 35164, "Load L1", Kind.AC),
        # 35165 reserved
        PowerS("load_p2", 35166, "Load L2", Kind.AC),
        # 35167 reserved
        PowerS("load_p3", 35168, "Load L3", Kind.AC),
        # 35169 reserved
        PowerS("backup_ptotal", 35170, "Back-up Load", Kind.UPS),
        # 35171 reserved
        PowerS("load_ptotal", 35172, "Load", Kind.AC),
        Integer("ups_load", 35173, "Ups Load", "%", Kind.UPS),
        Temp("temperature_air", 35174, "Inverter Temperature (Air)", Kind.AC),
        Temp("temperature_module", 35175, "Inverter Temperature (Module)"),
        Temp("temperature", 35176, "Inverter Temperature (Radiator)", Kind.AC),
        Integer("function_bit", 35177, "Function Bit"),
        Voltage("bus_voltage", 35178, "Bus Voltage", None),
        Voltage("nbus_voltage", 35179, "NBus Voltage", None),
        Voltage("vbattery1", 35180, "Battery Voltage", Kind.BAT),
        CurrentS("ibattery1", 35181, "Battery Current", Kind.BAT),
        Power4S("pbattery1", 35182, "Battery Power", Kind.BAT),
        Integer("battery_mode", 35184, "Battery Mode code", "", Kind.BAT),
        Enum2("battery_mode_label", 35184, BATTERY_MODES, "Battery Mode", Kind.BAT),
        Integer("warning_code", 35185, "Warning code"),
        Integer("safety_country", 35186, "Safety Country code", "", Kind.AC),
        Enum2("safety_country_label", 35186, SAFETY_COUNTRIES, "Safety Country", Kind.AC),
        Integer("work_mode", 35187, "Work Mode code"),
        Enum2("work_mode_label", 35187, WORK_MODES_ET, "Work Mode"),
        Integer("operation_mode", 35188, "Operation Mode code"),
        Long("error_codes", 35189, "Error Codes"),
        EnumBitmap4("errors", 35189, ERROR_CODES, "Errors"),
        Energy4("e_total", 35191, "Total PV Generation", Kind.PV),
        Energy4("e_day", 35193, "Today's PV Generation", Kind.PV),
        Energy4("e_total_exp", 35195, "Total Energy (export)", Kind.AC),
        Long("h_total", 35197, "Hours Total", "h", Kind.PV),
        Energy("e_day_exp", 35199, "Today Energy (export)", Kind.AC),
        Energy4("e_total_imp", 35200, "Total Energy (import)", Kind.AC),
        Energy("e_day_imp", 35202, "Today Energy (import)", Kind.AC),
        Energy4("e_load_total", 35203, "Total Load", Kind.AC),
        Energy("e_load_day", 35205, "Today Load", Kind.AC),
        Energy4("e_bat_charge_total", 35206, "Total Battery Charge", Kind.BAT),
        Energy("e_bat_charge_day", 35208, "Today Battery Charge", Kind.BAT),
        Energy4("e_bat_discharge_total", 35209, "Total Battery Discharge", Kind.BAT),
        Energy("e_bat_discharge_day", 35211, "Today Battery Discharge", Kind.BAT),
        Long("diagnose_result", 35220, "Diag Status Code"),
        EnumBitmap4("diagnose_result_label", 35220, DIAG_STATUS_CODES, "Diag Status"),
        # ppv1 + ppv2 + ppv3 + ppv4 + pbattery1 - active_power
        Calculated("house_consumption",
                   lambda data:
                   read_bytes4(data, 35105, 0) +
                   read_bytes4(data, 35109, 0) +
                   read_bytes4(data, 35113, 0) +
                   read_bytes4(data, 35117, 0) +
                   read_bytes4_signed(data, 35182) -
                   read_bytes2_signed(data, 35140),
                   "House Consumption", "W", Kind.AC),

        # Power4S("pbattery2", 35264, "Battery2 Power", Kind.BAT),
        # Integer("battery2_mode", 35266, "Battery2 Mode code", "", Kind.BAT),
        # Enum2("battery2_mode_label", 35184, BATTERY_MODES, "Battery2 Mode", Kind.BAT),
    )

    # Modbus registers from offset 0x9088 (37000)
    __all_sensors_battery: Tuple[Sensor, ...] = (
        Integer("battery_bms", 37000, "Battery BMS", "", Kind.BAT),
        Integer("battery_index", 37001, "Battery Index", "", Kind.BAT),
        Integer("battery_status", 37002, "Battery Status", "", Kind.BAT),
        Temp("battery_temperature", 37003, "Battery Temperature", Kind.BAT),
        Integer("battery_charge_limit", 37004, "Battery Charge Limit", "A", Kind.BAT),
        Integer("battery_discharge_limit", 37005, "Battery Discharge Limit", "A", Kind.BAT),
        Integer("battery_error_l", 37006, "Battery Error L", "", Kind.BAT),
        Integer("battery_soc", 37007, "Battery State of Charge", "%", Kind.BAT),
        Integer("battery_soh", 37008, "Battery State of Health", "%", Kind.BAT),
        Integer("battery_modules", 37009, "Battery Modules", "", Kind.BAT),
        Integer("battery_warning_l", 37010, "Battery Warning L", "", Kind.BAT),
        Integer("battery_protocol", 37011, "Battery Protocol", "", Kind.BAT),
        Integer("battery_error_h", 37012, "Battery Error H", "", Kind.BAT),
        EnumBitmap22("battery_error", 37012, 37006, BMS_ALARM_CODES, "Battery Error", Kind.BAT),
        Integer("battery_warning_h", 37013, "Battery Warning H", "", Kind.BAT),
        EnumBitmap22("battery_warning", 37013, 37010, BMS_WARNING_CODES, "Battery Warning", Kind.BAT),
        Integer("battery_sw_version", 37014, "Battery Software Version", "", Kind.BAT),
        Integer("battery_hw_version", 37015, "Battery Hardware Version", "", Kind.BAT),
        Integer("battery_max_cell_temp_id", 37016, "Battery Max Cell Temperature ID", "", Kind.BAT),
        Integer("battery_min_cell_temp_id", 37017, "Battery Min Cell Temperature ID", "", Kind.BAT),
        Integer("battery_max_cell_voltage_id", 37018, "Battery Max Cell Voltage ID", "", Kind.BAT),
        Integer("battery_min_cell_voltage_id", 37019, "Battery Min Cell Voltage ID", "", Kind.BAT),
        Temp("battery_max_cell_temp", 37020, "Battery Max Cell Temperature", Kind.BAT),
        Temp("battery_min_cell_temp", 37021, "Battery Min Cell Temperature", Kind.BAT),
        CellVoltage("battery_max_cell_voltage", 37022, "Battery Max Cell Voltage", Kind.BAT),
        CellVoltage("battery_min_cell_voltage", 37023, "Battery Min Cell Voltage", Kind.BAT),
        # Energy4("battery_total_charge", 37056, "Total Battery 1 Charge", Kind.BAT),
        # Energy4("battery_total_discharge", 37058, "Total Battery 1 Discharge", Kind.BAT),
        # String8("battery_sn", 37060, "Battery S/N", Kind.BAT),
    )

    # Modbus registers from offset 0x9858 (39000)
    __all_sensors_battery2: Tuple[Sensor, ...] = (
        Integer("battery2_status", 39000, "Battery 2 Status", "", Kind.BAT),
        Temp("battery2_temperature", 39001, "Battery 2 Temperature", Kind.BAT),
        Integer("battery2_charge_limit", 39002, "Battery 2 Charge Limit", "A", Kind.BAT),
        Integer("battery2_discharge_limit", 39003, "Battery 2 Discharge Limit", "A", Kind.BAT),
        Integer("battery2_error_l", 39004, "Battery 2 rror L", "", Kind.BAT),
        Integer("battery2_soc", 39005, "Battery 2 State of Charge", "%", Kind.BAT),
        Integer("battery2_soh", 39006, "Battery 2 State of Health", "%", Kind.BAT),
        Integer("battery2_modules", 39007, "Battery 2 Modules", "", Kind.BAT),
        Integer("battery2_warning_l", 39008, "Battery 2 Warning L", "", Kind.BAT),
        Integer("battery2_protocol", 39009, "Battery 2 Protocol", "", Kind.BAT),
        Integer("battery2_error_h", 39010, "Battery 2 Error H", "", Kind.BAT),
        EnumBitmap22("battery2_error", 39010, 39004, BMS_ALARM_CODES, "Battery 2 Error", Kind.BAT),
        Integer("battery2_warning_h", 39011, "Battery 2 Warning H", "", Kind.BAT),
        EnumBitmap22("battery2_warning", 39011, 39008, BMS_WARNING_CODES, "Battery 2 Warning", Kind.BAT),
        Integer("battery2_sw_version", 39012, "Battery 2 Software Version", "", Kind.BAT),
        Integer("battery2_hw_version", 39013, "Battery 2 Hardware Version", "", Kind.BAT),
        Integer("battery2_max_cell_temp_id", 39014, "Battery 2 Max Cell Temperature ID", "", Kind.BAT),
        Integer("battery2_min_cell_temp_id", 39015, "Battery 2 Min Cell Temperature ID", "", Kind.BAT),
        Integer("battery2_max_cell_voltage_id", 39016, "Battery 2 Max Cell Voltage ID", "", Kind.BAT),
        Integer("battery2_min_cell_voltage_id", 39017, "Battery 2 Min Cell Voltage ID", "", Kind.BAT),
        Temp("battery2_max_cell_temp", 39018, "Battery 2 Max Cell Temperature", Kind.BAT),
        Temp("battery2_min_cell_temp", 39019, "Battery 2 Min Cell Temperature", Kind.BAT),
        CellVoltage("battery2_max_cell_voltage", 39020, "Battery 2 Max Cell Voltage", Kind.BAT),
        CellVoltage("battery2_min_cell_voltage", 39021, "Battery 2 Min Cell Voltage", Kind.BAT),
        # Energy4("battery2_total_charge", 39054, "Total Battery 2 Charge", Kind.BAT),
        # Energy4("battery2_total_discharge", 39056, "Total Battery 2 Discharge", Kind.BAT),
        # String8("battery2_sn", 39058, "Battery 2 S/N", Kind.BAT),
    )

    # Inverter's meter data
    # Modbus registers from offset 0x8ca0 (36000)
    __all_sensors_meter: Tuple[Sensor, ...] = (
        Integer("commode", 36000, "Commode"),
        Integer("rssi", 36001, "RSSI"),
        Integer("manufacture_code", 36002, "Manufacture Code"),
        Integer("meter_test_status", 36003, "Meter Test Status"),  # 1: correct，2: reverse，3: incorrect，0: not checked
        Integer("meter_comm_status", 36004, "Meter Communication Status"),  # 1 OK, 0 NotOK
        PowerS("active_power1", 36005, "Active Power L1", Kind.GRID),
        PowerS("active_power2", 36006, "Active Power L2", Kind.GRID),
        PowerS("active_power3", 36007, "Active Power L3", Kind.GRID),
        PowerS("active_power_total", 36008, "Active Power Total", Kind.GRID),
        Reactive("reactive_power_total", 36009, "Reactive Power Total", Kind.GRID),
        Decimal("meter_power_factor1", 36010, 1000, "Meter Power Factor L1", "", Kind.GRID),
        Decimal("meter_power_factor2", 36011, 1000, "Meter Power Factor L2", "", Kind.GRID),
        Decimal("meter_power_factor3", 36012, 1000, "Meter Power Factor L3", "", Kind.GRID),
        Decimal("meter_power_factor", 36013, 1000, "Meter Power Factor", "", Kind.GRID),
        Frequency("meter_freq", 36014, "Meter Frequency", Kind.GRID),
        Float("meter_e_total_exp", 36015, 1000, "Meter Total Energy (export)", "kWh", Kind.GRID),
        Float("meter_e_total_imp", 36017, 1000, "Meter Total Energy (import)", "kWh", Kind.GRID),
        Power4S("meter_active_power1", 36019, "Meter Active Power L1", Kind.GRID),
        Power4S("meter_active_power2", 36021, "Meter Active Power L2", Kind.GRID),
        Power4S("meter_active_power3", 36023, "Meter Active Power L3", Kind.GRID),
        Power4S("meter_active_power_total", 36025, "Meter Active Power Total", Kind.GRID),
        Reactive4("meter_reactive_power1", 36027, "Meter Reactive Power L1", Kind.GRID),
        Reactive4("meter_reactive_power2", 36029, "Meter Reactive Power L2", Kind.GRID),
        Reactive4("meter_reactive_power3", 36031, "Meter Reactive Power L2", Kind.GRID),
        Reactive4("meter_reactive_power_total", 36033, "Meter Reactive Power Total", Kind.GRID),
        Apparent4("meter_apparent_power1", 36035, "Meter Apparent Power L1", Kind.GRID),
        Apparent4("meter_apparent_power2", 36037, "Meter Apparent Power L2", Kind.GRID),
        Apparent4("meter_apparent_power3", 36039, "Meter Apparent Power L3", Kind.GRID),
        Apparent4("meter_apparent_power_total", 36041, "Meter Apparent Power Total", Kind.GRID),
        Integer("meter_type", 36043, "Meter Type", "", Kind.GRID),  # (0: Single phase, 1: 3P3W, 2: 3P4W, 3: HomeKit)
        Integer("meter_sw_version", 36044, "Meter Software Version", "", Kind.GRID),

        # Sensors added in some ARM fw update (or platform 745/753), read when flag _has_meter_extended is on
        Power4S("meter2_active_power", 36045, "Meter 2 Active Power", Kind.GRID),
        Float("meter2_e_total_exp", 36047, 1000, "Meter 2 Total Energy (export)", "kWh", Kind.GRID),
        Float("meter2_e_total_imp", 36049, 1000, "Meter 2 Total Energy (import)", "kWh", Kind.GRID),
        Integer("meter2_comm_status", 36051, "Meter 2 Communication Status"),
        Voltage("meter_voltage1", 36052, "Meter L1 Voltage", Kind.GRID),
        Voltage("meter_voltage2", 36053, "Meter L2 Voltage", Kind.GRID),
        Voltage("meter_voltage3", 36054, "Meter L3 Voltage", Kind.GRID),
        Current("meter_current1", 36055, "Meter L1 Current", Kind.GRID),
        Current("meter_current2", 36056, "Meter L2 Current", Kind.GRID),
        Current("meter_current3", 36057, "Meter L3 Current", Kind.GRID),

        Energy8("meter_e_total_exp1", 36092, "Meter Total Energy (export) L1", Kind.GRID),
        Energy8("meter_e_total_exp2", 36096, "Meter Total Energy (export) L2", Kind.GRID),
        Energy8("meter_e_total_exp3", 36100, "Meter Total Energy (export) L3", Kind.GRID),
        Energy8("meter_e_total_exp", 36104, "Meter Total Energy (export)", Kind.GRID),
        Energy8("meter_e_total_imp1", 36108, "Meter Total Energy (import) L1", Kind.GRID),
        Energy8("meter_e_total_imp2", 36112, "Meter Total Energy (import) L2", Kind.GRID),
        Energy8("meter_e_total_imp3", 36116, "Meter Total Energy (import) L3", Kind.GRID),
        Energy8("meter_e_total_imp", 36120, "Meter Total Energy (import)", Kind.GRID),
    )

    # Inverter's MPPT data
    # Modbus registers from offset 0x89e5 (35301)
    __all_sensors_mppt: Tuple[Sensor, ...] = (
        Power4("ppv_total", 35301, "PV Power Total", Kind.PV),
        Integer("pv_channel", 35303, "PV Channel", "", Kind.PV),
        Voltage("vpv5", 35304, "PV5 Voltage", Kind.PV),
        Current("ipv5", 35305, "PV5 Current", Kind.PV),
        Voltage("vpv6", 35306, "PV6 Voltage", Kind.PV),
        Current("ipv6", 35307, "PV6 Current", Kind.PV),
        Voltage("vpv7", 35308, "PV7 Voltage", Kind.PV),
        Current("ipv7", 35309, "PV7 Current", Kind.PV),
        Voltage("vpv8", 35310, "PV8 Voltage", Kind.PV),
        Current("ipv8", 35311, "PV8 Current", Kind.PV),
        Voltage("vpv9", 35312, "PV9 Voltage", Kind.PV),
        Current("ipv9", 35313, "PV9 Current", Kind.PV),
        Voltage("vpv10", 35314, "PV10 Voltage", Kind.PV),
        Current("ipv10", 35315, "PV10 Current", Kind.PV),
        Voltage("vpv11", 35316, "PV11 Voltage", Kind.PV),
        Current("ipv11", 35317, "PV11 Current", Kind.PV),
        Voltage("vpv12", 35318, "PV12 Voltage", Kind.PV),
        Current("ipv12", 35319, "PV12 Current", Kind.PV),
        Voltage("vpv13", 35320, "PV13 Voltage", Kind.PV),
        Current("ipv13", 35321, "PV13 Current", Kind.PV),
        Voltage("vpv14", 35322, "PV14 Voltage", Kind.PV),
        Current("ipv14", 35323, "PV14 Current", Kind.PV),
        Voltage("vpv15", 35324, "PV15 Voltage", Kind.PV),
        Current("ipv15", 35325, "PV15 Current", Kind.PV),
        Voltage("vpv16", 35326, "PV16 Voltage", Kind.PV),
        Current("ipv16", 35327, "PV16 Current", Kind.PV),
        # 35328 Warning Message
        # 35330 Grid10minAvgVoltR
        # 35331 Grid10minAvgVoltS
        # 35332 Grid10minAvgVoltT
        # 35333 Error Message Extend
        # 35335 Warning Message Extend
        Power("pmppt1", 35337, "MPPT1 Power", Kind.PV),
        Power("pmppt2", 35338, "MPPT2 Power", Kind.PV),
        Power("pmppt3", 35339, "MPPT3 Power", Kind.PV),
        Power("pmppt4", 35340, "MPPT4 Power", Kind.PV),
        Power("pmppt5", 35341, "MPPT5 Power", Kind.PV),
        Power("pmppt6", 35342, "MPPT6 Power", Kind.PV),
        Power("pmppt7", 35343, "MPPT7 Power", Kind.PV),
        Power("pmppt8", 35344, "MPPT8 Power", Kind.PV),
        Current("imppt1", 35345, "MPPT1 Current", Kind.PV),
        Current("imppt2", 35346, "MPPT2 Current", Kind.PV),
        Current("imppt3", 35347, "MPPT3 Current", Kind.PV),
        Current("imppt4", 35348, "MPPT4 Current", Kind.PV),
        Current("imppt5", 35349, "MPPT5 Current", Kind.PV),
        Current("imppt6", 35350, "MPPT6 Current", Kind.PV),
        Current("imppt7", 35351, "MPPT7 Current", Kind.PV),
        Current("imppt8", 35352, "MPPT8 Current", Kind.PV),
        Reactive4("reactive_power1", 35353, "Reactive Power L1", Kind.GRID),
        Reactive4("reactive_power2", 35355, "Reactive Power L2", Kind.GRID),
        Reactive4("reactive_power3", 35357, "Reactive Power L3", Kind.GRID),
        Apparent4("apparent_power1", 35359, "Apparent Power L1", Kind.GRID),
        Apparent4("apparent_power2", 35361, "Apparent Power L2", Kind.GRID),
        Apparent4("apparent_power3", 35363, "Apparent Power L3", Kind.GRID),
    )

    # Modbus registers of inverter settings, offsets are modbus register addresses
    __all_settings: Tuple[Sensor, ...] = (
        Integer("comm_address", 45127, "Communication Address", ""),
        Long("modbus_baud_rate", 45132, "Modbus Baud rate", ""),
        Timestamp("time", 45200, "Inverter time"),

        Integer("sensitivity_check", 45246, "Sensitivity Check Mode", "", Kind.AC),
        Integer("cold_start", 45248, "Cold Start", "", Kind.AC),
        Integer("shadow_scan", 45251, "Shadow Scan", "", Kind.PV),
        Integer("backup_supply", 45252, "Backup Supply", "", Kind.UPS),
        Integer("unbalanced_output", 45264, "Unbalanced Output", "", Kind.AC),
        Integer("pen_relay", 45288, "PE-N Relay", "", Kind.AC),

        Integer("battery_capacity", 45350, "Battery Capacity", "Ah", Kind.BAT),
        Integer("battery_modules", 45351, "Battery Modules", "", Kind.BAT),
        Voltage("battery_charge_voltage", 45352, "Battery Charge Voltage", Kind.BAT),
        Current("battery_charge_current", 45353, "Battery Charge Current", Kind.BAT),
        Voltage("battery_discharge_voltage", 45354, "Battery Discharge Voltage", Kind.BAT),
        Current("battery_discharge_current", 45355, "Battery Discharge Current", Kind.BAT),
        Integer("battery_discharge_depth", 45356, "Battery Discharge Depth", "%", Kind.BAT),
        Voltage("battery_discharge_voltage_offline", 45357, "Battery Discharge Voltage (off-line)", Kind.BAT),
        Integer("battery_discharge_depth_offline", 45358, "Battery Discharge Depth (off-line)", "%", Kind.BAT),

        Decimal("power_factor", 45482, 100, "Power Factor"),

        Integer("work_mode", 47000, "Work Mode", "", Kind.AC),
        Integer("dred", 47010, "DRED/Remote Shutdown", "", Kind.AC),

        Integer("meter_target_power_offset", 47120, "Meter Target Power Offset", "W", Kind.AC),

        Integer("battery_soc_protection", 47500, "Battery SoC Protection", "", Kind.BAT),

        Integer("grid_export", 47509, "Grid Export Enabled", "", Kind.GRID),
        Integer("grid_export_limit", 47510, "Grid Export Limit", "W", Kind.GRID),

        Integer("battery_protocol_code", 47514, "Battery Protocol Code", "", Kind.BAT),

        EcoModeV1("eco_mode_1", 47515, "Eco Mode Group 1"),
        ByteH("eco_mode_1_switch", 47518, "Eco Mode Group 1 Switch"),
        EcoModeV1("eco_mode_2", 47519, "Eco Mode Group 2"),
        ByteH("eco_mode_2_switch", 47522, "Eco Mode Group 2 Switch"),
        EcoModeV1("eco_mode_3", 47523, "Eco Mode Group 3"),
        ByteH("eco_mode_3_switch", 47526, "Eco Mode Group 3 Switch"),
        EcoModeV1("eco_mode_4", 47527, "Eco Mode Group 4"),
        ByteH("eco_mode_4_switch", 47530, "Eco Mode Group 4 Switch"),

        # Direct BMS communication for EMS Control
        Integer("bms_version", 47900, "BMS Version"),
        Integer("bms_bat_modules", 47901, "BMS Battery Modules"),
        # Real time read from BMS
        Voltage("bms_bat_charge_v_max", 47902, "BMS Battery Charge Voltage (max)", Kind.BMS),
        Current("bms_bat_charge_i_max", 47903, "BMS Battery Charge Current (max)", Kind.BMS),
        Voltage("bms_bat_discharge_v_min", 47904, "BMS min. Battery Discharge Voltage (min)", Kind.BMS),
        Current("bms_bat_discharge_i_max", 47905, "BMS max. Battery Discharge Current (max)", Kind.BMS),
        Voltage("bms_bat_voltage", 47906, "BMS Battery Voltage", Kind.BMS),
        Current("bms_bat_current", 47907, "BMS Battery Current", Kind.BMS),
        #
        Integer("bms_bat_soc", 47908, "BMS Battery State of Charge", "%", Kind.BMS),
        Integer("bms_bat_soh", 47909, "BMS Battery State of Health", "%", Kind.BMS),
        Temp("bms_bat_temperature", 47910, "BMS Battery Temperature", Kind.BMS),
        Long("bms_bat_warning-code", 47911, "BMS Battery Warning Code"),
        # Reserved
        Long("bms_bat_alarm-code", 47913, "BMS Battery Alarm Code"),
        Integer("bms_status", 47915, "BMS Status"),
        Integer("bms_comm_loss_disable", 47916, "BMS Communication Loss Disable"),
        # RW settings of BMS voltage rate
        Integer("bms_battery_string_rate_v", 47917, "BMS Battery String Rate Voltage"),

        # Direct BMS communication for EMS Control
        Integer("bms2_version", 47918, "BMS2 Version"),
        Integer("bms2_bat_modules", 47919, "BMS2 Battery Modules"),
        # Real time read from BMS
        Voltage("bms2_bat_charge_v_max", 47920, "BMS2 Battery Charge Voltage (max)", Kind.BMS),
        Current("bms2_bat_charge_i_max", 47921, "BMS2 Battery Charge Current (max)", Kind.BMS),
        Voltage("bms2_bat_discharge_v_min", 47922, "BMS2 min. Battery Discharge Voltage (min)", Kind.BMS),
        Current("bms2_bat_discharge_i_max", 47923, "BMS2 max. Battery Discharge Current (max)", Kind.BMS),
        Voltage("bms2_bat_voltage", 47924, "BMS2 Battery Voltage", Kind.BMS),
        Current("bms2_bat_current", 47925, "BMS2 Battery Current", Kind.BMS),
        #
        Integer("bms2_bat_soc", 47926, "BMS2 Battery State of Charge", "%", Kind.BMS),
        Integer("bms2_bat_soh", 47927, "BMS2 Battery State of Health", "%", Kind.BMS),
        Temp("bms2_bat_temperature", 47928, "BMS2 Battery Temperature", Kind.BMS),
        Long("bms2_bat_warning-code", 47929, "BMS2 Battery Warning Code"),
        # Reserved
        Long("bms2_bat_alarm-code", 47931, "BMS2 Battery Alarm Code"),
        Integer("bms2_status", 47933, "BMS2 Status"),
        Integer("bms2_comm_loss_disable", 47934, "BMS2 Communication Loss Disable"),
        # RW settings of BMS voltage rate
        Integer("bms2_battery_string_rate_v", 47935, "BMS2 Battery String Rate Voltage"),

    )

    # Settings added in ARM firmware 19
    __settings_arm_fw_19: Tuple[Sensor, ...] = (
        Integer("fast_charging", 47545, "Fast Charging Enabled", "", Kind.BAT),
        Integer("fast_charging_soc", 47546, "Fast Charging SoC", "%", Kind.BAT),
        EcoModeV2("eco_mode_1", 47547, "Eco Mode Group 1"),
        ByteH("eco_mode_1_switch", 47549, "Eco Mode Group 1 Switch"),
        EcoModeV2("eco_mode_2", 47553, "Eco Mode Group 2"),
        ByteH("eco_mode_2_switch", 47555, "Eco Mode Group 2 Switch"),
        EcoModeV2("eco_mode_3", 47559, "Eco Mode Group 3"),
        ByteH("eco_mode_3_switch", 47561, "Eco Mode Group 3 Switch"),
        EcoModeV2("eco_mode_4", 47565, "Eco Mode Group 4"),
        ByteH("eco_mode_4_switch", 47567, "Eco Mode Group 4 Switch"),

        Integer("load_control_mode", 47595, "Load Control Mode", "", Kind.AC),
        Integer("load_control_switch", 47596, "Load Control Switch", "", Kind.AC),
        Integer("load_control_soc", 47597, "Load Control SoC", "", Kind.AC),
        Integer("hardware_feed_power", 47599, "Hardware Feed Power"),

        Integer("fast_charging_power", 47603, "Fast Charging Power", "%", Kind.BAT),
    )

    # Settings added in ARM firmware 22
    __settings_arm_fw_22: Tuple[Sensor, ...] = (
        Long("peak_shaving_power_limit", 47542, "Peak Shaving Power Limit"),
        Integer("peak_shaving_soc", 47544, "Peak Shaving SoC"),
        # EcoModeV2("eco_modeV2_5", 47571, "Eco Mode Version 2 Power Group 5"),
        # EcoModeV2("eco_modeV2_6", 47577, "Eco Mode Version 2 Power Group 6"),
        # EcoModeV2("eco_modeV2_7", 47583, "Eco Mode Version 2 Power Group 7"),
        PeakShavingMode("peak_shaving_mode", 47589, "Peak Shaving Mode"),

        Integer("dod_holding", 47602, "DoD Holding", "", Kind.BAT),
        Integer("backup_mode_enable", 47605, "Backup Mode Switch"),
        Integer("max_charge_power", 47606, "Max Charge Power"),
        Integer("smart_charging_enable", 47609, "Smart Charging Mode Switch"),
        Integer("eco_mode_enable", 47612, "Eco Mode Switch"),
    )

    def __init__(self, host: str, port: int, comm_addr: int = 0, timeout: int = 1, retries: int = 3):
        super().__init__(host, port, comm_addr if comm_addr else 0xf7, timeout, retries)
        self._READ_DEVICE_VERSION_INFO: ProtocolCommand = self._read_command(0x88b8, 0x0021)
        self._READ_RUNNING_DATA: ProtocolCommand = self._read_command(0x891c, 0x007d)
        self._READ_METER_DATA: ProtocolCommand = self._read_command(0x8ca0, 0x2d)
        self._READ_METER_DATA_EXTENDED: ProtocolCommand = self._read_command(0x8ca0, 0x3a)
        self._READ_METER_DATA_EXTENDED2: ProtocolCommand = self._read_command(0x8ca0, 0x7d)
        self._READ_BATTERY_INFO: ProtocolCommand = self._read_command(0x9088, 0x0018)
        self._READ_BATTERY2_INFO: ProtocolCommand = self._read_command(0x9858, 0x0016)
        self._READ_MPPT_DATA: ProtocolCommand = self._read_command(0x89e5, 0x3d)
        self._has_eco_mode_v2: bool = True
        self._has_peak_shaving: bool = True
        self._has_battery: bool = True
        self._has_battery2: bool = False
        self._has_meter_extended: bool = False
        self._has_meter_extended2: bool = False
        self._has_mppt: bool = False
        self._sensors = self.__all_sensors
        self._sensors_battery = self.__all_sensors_battery
        self._sensors_battery2 = self.__all_sensors_battery2
        self._sensors_meter = self.__all_sensors_meter
        self._sensors_mppt = self.__all_sensors_mppt
        self._settings: dict[str, Sensor] = {s.id_: s for s in self.__all_settings}

    @staticmethod
    def _single_phase_only(s: Sensor) -> bool:
        """Filter to exclude phase2/3 sensors on single phase inverters"""
        return not ((s.id_.endswith('2') or s.id_.endswith('3')) and 'pv' not in s.id_)

    @staticmethod
    def _not_extended_meter(s: Sensor) -> bool:
        """Filter to exclude extended meter sensors"""
        return s.offset < 36045

    @staticmethod
    def _not_extended_meter2(s: Sensor) -> bool:
        """Filter to exclude extended meter sensors"""
        return s.offset < 36058

    async def read_device_info(self):
        response = await self._read_from_socket(self._READ_DEVICE_VERSION_INFO)
        response = response.response_data()
        # Modbus registers from 35000 - 35032
        self.modbus_version = read_unsigned_int(response, 0)
        self.rated_power = read_unsigned_int(response, 2)
        self.ac_output_type = read_unsigned_int(response, 4)  # 0: 1-phase, 1: 3-phase (4 wire), 2: 3-phase (3 wire)
        self.serial_number = self._decode(response[6:22])  # 35003 - 350010
        self.model_name = self._decode(response[22:32])  # 35011 - 35015
        self.dsp1_version = read_unsigned_int(response, 32)  # 35016
        self.dsp2_version = read_unsigned_int(response, 34)  # 35017
        self.dsp_svn_version = read_unsigned_int(response, 36)  # 35018
        self.arm_version = read_unsigned_int(response, 38)  # 35019
        self.arm_svn_version = read_unsigned_int(response, 40)  # 35020
        self.firmware = self._decode(response[42:54])  # 35021 - 35027
        self.arm_firmware = self._decode(response[54:66])  # 35027 - 35032

        if not is_4_mppt(self) and self.rated_power < 15000:
            # This inverter does not have 4 MPPTs or PV strings
            self._sensors = tuple(filter(lambda s: not ('pv4' in s.id_), self._sensors))
            self._sensors = tuple(filter(lambda s: not ('pv3' in s.id_), self._sensors))

        if is_single_phase(self):
            # this is single phase inverter, filter out all L2 and L3 sensors
            self._sensors = tuple(filter(self._single_phase_only, self._sensors))
            self._sensors_meter = tuple(filter(self._single_phase_only, self._sensors_meter))

        if is_2_battery(self) or self.rated_power >= 25000:
            self._has_battery2 = True

        if is_745_platform(self) or self.rated_power >= 15000:
            self._has_mppt = True
            self._has_meter_extended = True
            self._has_meter_extended2 = True
        else:
            self._sensors_meter = tuple(filter(self._not_extended_meter, self._sensors_meter))

        # Check and add EcoModeV2 settings added in (ETU fw 19)
        try:
            await self._read_from_socket(self._read_command(47547, 6))
            self._settings.update({s.id_: s for s in self.__settings_arm_fw_19})
        except RequestRejectedException as ex:
            if ex.message == ILLEGAL_DATA_ADDRESS:
                logger.debug("EcoModeV2 settings not supported, switching to EcoModeV1.")
                self._has_eco_mode_v2 = False
        except RequestFailedException:
            logger.debug("Cannot read EcoModeV2 settings, switching to EcoModeV1.")
            self._has_eco_mode_v2 = False

        # Check and add Peak Shaving settings added in (ETU fw 22)
        try:
            await self._read_from_socket(self._read_command(47589, 6))
            self._settings.update({s.id_: s for s in self.__settings_arm_fw_22})
        except RequestRejectedException as ex:
            if ex.message == ILLEGAL_DATA_ADDRESS:
                logger.debug("PeakShaving setting not supported, disabling it.")
                self._has_peak_shaving = False
        except RequestFailedException:
            logger.debug("Cannot read _has_peak_shaving settings, disabling it.")
            self._has_peak_shaving = False

    async def read_runtime_data(self) -> Dict[str, Any]:
        response = await self._read_from_socket(self._READ_RUNNING_DATA)
        data = self._map_response(response, self._sensors)

        self._has_battery = data.get('battery_mode', 0) != 0
        if self._has_battery:
            try:
                response = await self._read_from_socket(self._READ_BATTERY_INFO)
                data.update(self._map_response(response, self._sensors_battery))
            except RequestRejectedException as ex:
                if ex.message == ILLEGAL_DATA_ADDRESS:
                    logger.info("Battery values not supported, disabling further attempts.")
                    self._has_battery = False
                else:
                    raise ex
        if self._has_battery2:
            try:
                response = await self._read_from_socket(self._READ_BATTERY2_INFO)
                data.update(
                    self._map_response(response, self._sensors_battery2))
            except RequestRejectedException as ex:
                if ex.message == ILLEGAL_DATA_ADDRESS:
                    logger.info("Battery 2 values not supported, disabling further attempts.")
                    self._has_battery2 = False
                else:
                    raise ex

        if self._has_meter_extended2:
            try:
                response = await self._read_from_socket(self._READ_METER_DATA_EXTENDED2)
                data.update(self._map_response(response, self._sensors_meter))
            except RequestRejectedException as ex:
                if ex.message == ILLEGAL_DATA_ADDRESS:
                    logger.info("Extended meter values not supported, disabling further attempts.")
                    self._has_meter_extended2 = False
                    self._sensors_meter = tuple(filter(self._not_extended_meter2, self._sensors_meter))
                    response = await self._read_from_socket(self._READ_METER_DATA_EXTENDED)
                    data.update(
                        self._map_response(response, self._sensors_meter))
                else:
                    raise ex
        elif self._has_meter_extended:
            try:
                response = await self._read_from_socket(self._READ_METER_DATA_EXTENDED)
                data.update(self._map_response(response, self._sensors_meter))
            except RequestRejectedException as ex:
                if ex.message == ILLEGAL_DATA_ADDRESS:
                    logger.info("Extended meter values not supported, disabling further attempts.")
                    self._has_meter_extended = False
                    self._sensors_meter = tuple(filter(self._not_extended_meter, self._sensors_meter))
                    response = await self._read_from_socket(self._READ_METER_DATA)
                    data.update(
                        self._map_response(response, self._sensors_meter))
                else:
                    raise ex
        else:
            response = await self._read_from_socket(self._READ_METER_DATA)
            data.update(self._map_response(response, self._sensors_meter))

        if self._has_mppt:
            try:
                response = await self._read_from_socket(self._READ_MPPT_DATA)
                data.update(self._map_response(response, self._sensors_mppt))
            except RequestRejectedException as ex:
                if ex.message == ILLEGAL_DATA_ADDRESS:
                    logger.info("MPPT values not supported, disabling further attempts.")
                    self._has_mppt = False
                else:
                    raise ex

        return data

    async def read_setting(self, setting_id: str) -> Any:
        setting = self._settings.get(setting_id)
        if setting:
            return await self._read_setting(setting)
        else:
            if setting_id.startswith("modbus"):
                response = await self._read_from_socket(self._read_command(int(setting_id[7:]), 1))
                return int.from_bytes(response.read(2), byteorder="big", signed=True)
            else:
                raise ValueError(f'Unknown setting "{setting_id}"')

    async def _read_setting(self, setting: Sensor) -> Any:
        try:
            count = (setting.size_ + (setting.size_ % 2)) // 2
            response = await self._read_from_socket(self._read_command(setting.offset, count))
            return setting.read_value(response)
        except RequestRejectedException as ex:
            if ex.message == ILLEGAL_DATA_ADDRESS:
                logger.debug("Unsupported setting %s", setting.id_)
                self._settings.pop(setting.id_, None)
                raise ValueError(f'Unknown setting "{setting.id_}"')
            return None

    async def write_setting(self, setting_id: str, value: Any):
        setting = self._settings.get(setting_id)
        if setting:
            await self._write_setting(setting, value)
        else:
            if setting_id.startswith("modbus"):
                await self._read_from_socket(self._write_command(int(setting_id[7:]), int(value)))
            else:
                raise ValueError(f'Unknown setting "{setting_id}"')

    async def _write_setting(self, setting: Sensor, value: Any):
        if setting.size_ == 1:
            # modbus can address/store only 16 bit values, read the other 8 bytes
            response = await self._read_from_socket(self._read_command(setting.offset, 1))
            raw_value = setting.encode_value(value, response.response_data()[0:2])
        else:
            raw_value = setting.encode_value(value)
        if len(raw_value) <= 2:
            value = int.from_bytes(raw_value, byteorder="big", signed=True)
            await self._read_from_socket(self._write_command(setting.offset, value))
        else:
            await self._read_from_socket(self._write_multi_command(setting.offset, raw_value))

    async def read_settings_data(self) -> Dict[str, Any]:
        data = {}
        for setting in self.settings():
            try:
                value = await self.read_setting(setting.id_)
                data[setting.id_] = value
            except (ValueError, RequestFailedException):
                logger.exception("Error reading setting %s.", setting.id_)
                data[setting.id_] = None
        return data

    async def get_grid_export_limit(self) -> int:
        return await self.read_setting('grid_export_limit')

    async def set_grid_export_limit(self, export_limit: int) -> None:
        if export_limit >= 0:
            await self.write_setting('grid_export_limit', export_limit)

    async def get_operation_modes(self, include_emulated: bool) -> Tuple[OperationMode, ...]:
        result = [e for e in OperationMode]
        if not self._has_peak_shaving:
            result.remove(OperationMode.PEAK_SHAVING)
        if not is_745_platform(self):
            result.remove(OperationMode.SELF_USE)
        if not include_emulated:
            result.remove(OperationMode.ECO_CHARGE)
            result.remove(OperationMode.ECO_DISCHARGE)
        return tuple(result)

    async def get_operation_mode(self) -> OperationMode | None:
        mode_id = await self.read_setting('work_mode')
        try:
            mode = OperationMode(mode_id)
        except ValueError:
            logger.debug("Unknown work_mode value %s", mode_id)
            return None
        if OperationMode.ECO != mode:
            return mode
        eco_mode = await self.read_setting('eco_mode_1')
        if eco_mode.is_eco_charge_mode():
            return OperationMode.ECO_CHARGE
        elif eco_mode.is_eco_discharge_mode():
            return OperationMode.ECO_DISCHARGE
        else:
            return OperationMode.ECO

    async def set_operation_mode(self, operation_mode: OperationMode, eco_mode_power: int = 100,
                                 eco_mode_soc: int = 100) -> None:
        if operation_mode == OperationMode.GENERAL:
            await self.write_setting('work_mode', 0)
            await self._set_offline(False)
            await self._clear_battery_mode_param()
        elif operation_mode == OperationMode.OFF_GRID:
            await self.write_setting('work_mode', 1)
            await self._set_offline(True)
            await self.write_setting('backup_supply', 1)
            await self.write_setting('cold_start', 4)
            await self._clear_battery_mode_param()
        elif operation_mode == OperationMode.BACKUP:
            await self.write_setting('work_mode', 2)
            await self._set_offline(False)
            await self._clear_battery_mode_param()
        elif operation_mode == OperationMode.ECO:
            await self.write_setting('work_mode', 3)
            await self._set_offline(False)
        elif operation_mode == OperationMode.PEAK_SHAVING:
            await self.write_setting('work_mode', 4)
            await self._set_offline(False)
            await self._clear_battery_mode_param()
        elif operation_mode == OperationMode.SELF_USE:
            await self.write_setting('work_mode', 5)
            await self._set_offline(False)
            await self._clear_battery_mode_param()
        elif operation_mode in (OperationMode.ECO_CHARGE, OperationMode.ECO_DISCHARGE):
            if eco_mode_power < 0 or eco_mode_power > 100:
                raise ValueError()
            if eco_mode_soc < 0 or eco_mode_soc > 100:
                raise ValueError()

            eco_mode: EcoMode | Sensor = self._settings.get('eco_mode_1')
            # Load the current values to try to detect schedule type
            try:
                await self._read_setting(eco_mode)
            except ValueError:
                pass
            eco_mode.set_schedule_type(ScheduleType.ECO_MODE, is_745_platform(self))
            if operation_mode == OperationMode.ECO_CHARGE:
                await self.write_setting('eco_mode_1', eco_mode.encode_charge(eco_mode_power, eco_mode_soc))
            else:
                await self.write_setting('eco_mode_1', eco_mode.encode_discharge(eco_mode_power))
            await self.write_setting('eco_mode_2_switch', 0)
            await self.write_setting('eco_mode_3_switch', 0)
            await self.write_setting('eco_mode_4_switch', 0)
            await self.write_setting('work_mode', 3)
            await self._set_offline(False)

    async def get_ongrid_battery_dod(self) -> int:
        return 100 - await self.read_setting('battery_discharge_depth')

    async def set_ongrid_battery_dod(self, dod: int) -> None:
        if 0 <= dod <= 100:
            await self.write_setting('battery_discharge_depth', 100 - dod)

    def sensors(self) -> Tuple[Sensor, ...]:
        result = self._sensors + self._sensors_meter
        if self._has_battery:
            result = result + self._sensors_battery
        if self._has_battery2:
            result = result + self._sensors_battery2
        if self._has_mppt:
            result = result + self._sensors_mppt
        return result

    def settings(self) -> Tuple[Sensor, ...]:
        return tuple(self._settings.values())

    async def _clear_battery_mode_param(self) -> None:
        await self._read_from_socket(self._write_command(0xb9ad, 1))

    async def _set_offline(self, mode: bool) -> None:
        value = bytes.fromhex('00070000') if mode else bytes.fromhex('00010000')
        await self._read_from_socket(self._write_multi_command(0xb997, value))



from unittest import TestCase

from goodwe.modbus import *


class TestModbus(TestCase):

    def assert_rtu_response_ok(self, response: str, cmd: int, offset: int, value: int):
        self.assertTrue(validate_modbus_rtu_response(bytes.fromhex(response), cmd, offset, value))

    def assert_rtu_response_fail(self, response: str, cmd: int, offset: int, value: int):
        self.assertFalse(validate_modbus_rtu_response(bytes.fromhex(response), cmd, offset, value))

    def assert_rtu_response_partial(self, response: str, cmd: int, offset: int, value: int):
        self.assertRaises(PartialResponseException,
                          lambda: validate_modbus_rtu_response(bytes.fromhex(response), cmd, offset, value))

    def assert_rtu_response_rejected(self, response: str, cmd: int, offset: int, value: int):
        self.assertRaises(RequestRejectedException,
                          lambda: validate_modbus_rtu_response(bytes.fromhex(response), cmd, offset, value))

    def test_create_modbus_rtu_request(self):
        request = create_modbus_rtu_request(0x11, 0x3, 0x006b, 0x0003)
        self.assertEqual('1103006b00037687', request.hex())

        request = create_modbus_rtu_request(0xf7, 0x3, 0x88b8, 0x0021)
        self.assertEqual('f70388b800213ac1', request.hex())
        request = create_modbus_rtu_request(0xf7, 0x6, 0x88b8, 0x00FF)
        self.assertEqual('f70688b800ff7699', request.hex())
        request = create_modbus_rtu_request(0xf7, 0x6, 0x88b8, -1)
        self.assertEqual('f70688b8ffff3769', request.hex())

    def test_create_modbus_rtu_multi_request(self):
        request = create_modbus_rtu_multi_request(0xf7, 0x10, 0x88b8, b'\x01\x02\x03\x04\x05\x06')
        self.assertEqual('f71088b8000306010203040506102e', request.hex())

    def test_validate_modbus_rtu_read_response(self):
        self.assert_rtu_response_ok('aa55f7030401020304cd33', 0x03, 0x0401, 2)
        # some garbage after response end
        self.assert_rtu_response_ok('aa55f7030401020304cd33ffffff', 0x03, 0x0401, 2)
        # length too short
        self.assert_rtu_response_partial('aa55f7030401020304', 0x03, 0x0401, 2)
        # wrong checksum
        self.assert_rtu_response_fail('aa55f70304010203043346', 0x03, 0x0401, 2)
        # failure code
        self.assert_rtu_response_rejected('aa55f783040102030405b35e', 0x03, 0x0401, 2)
        # unexpected message length
        self.assert_rtu_response_fail('aa55f70306010203040506b417', 0x03, 0x0401, 2)

    def test_validate_modbus_rtu_write_response(self):
        self.assert_rtu_response_ok('aa55f706b12c00147ba6', 0x06, 0xb12c, 0x14)
        self.assert_rtu_response_ok('aa55f706b12cffff7a19', 0x06, 0xb12c, -1)
        # some garbage after response end
        self.assert_rtu_response_ok('aa55f706b12c00147ba6ffffff', 0x06, 0xb12c, 0x14)
        # length too short
        self.assert_rtu_response_fail('aa55f7066b12', 0x06, 0xb12c, 0x14)
        # wrong checksum
        self.assert_rtu_response_fail('aa55f706b12c00147ba7', 0x06, 0xb12c, 0x14)
        # wrong value written
        self.assert_rtu_response_fail('aa55f706b12c0012fba4', 0x06, 0xb12c, 0x14)

    def test_validate_modbus_rtu_write_multi_response(self):
        self.assert_rtu_response_ok('aa55f71088b800033f1b', 0x10, 0x88b8, 0x03)
        # some garbage after response end
        self.assert_rtu_response_ok('aa55f71088b800033f1bffffff', 0x10, 0x88b8, 0x03)
        # length too short
        self.assert_rtu_response_fail('aa55f71088b8', 0x10, 0x88b8, 0x03)
        # wrong checksum
        self.assert_rtu_response_fail('aa55f71088b800033f1c', 0x10, 0x88b8, 0x03)
        # wrong value written
        self.assert_rtu_response_fail('aa55f71088b80001beda', 0x10, 0x88b8, 0x03)

    def assert_tcp_response_ok(self, response: str, cmd: int, offset: int, value: int):
        self.assertTrue(validate_modbus_tcp_response(bytes.fromhex(response), cmd, offset, value))

    def assert_tcp_response_fail(self, response: str, cmd: int, offset: int, value: int):
        self.assertFalse(validate_modbus_tcp_response(bytes.fromhex(response), cmd, offset, value))

    def assert_tcp_response_partial(self, response: str, cmd: int, offset: int, value: int):
        self.assertRaises(PartialResponseException,
                          lambda: validate_modbus_tcp_response(bytes.fromhex(response), cmd, offset, value))

    def assert_tcp_response_rejected(self, response: str, cmd: int, offset: int, value: int):
        self.assertRaises(RequestRejectedException,
                          lambda: validate_modbus_tcp_response(bytes.fromhex(response), cmd, offset, value))

    def test_create_modbus_tcp_request(self):
        request = create_modbus_tcp_request(0x11, 0x3, 0x006b, 0x0003)
        self.assertEqual('0001000000061103006b0003', request.hex())
        request = create_modbus_tcp_request(180, 0x3, 331, 2)
        self.assertEqual('000100000006b403014b0002', request.hex())

        request = create_modbus_tcp_request(0xf7, 0x3, 0x88b8, 0x0021)
        self.assertEqual('000100000006f70388b80021', request.hex())
        request = create_modbus_tcp_request(0xf7, 0x6, 0x88b8, 0x00FF)
        self.assertEqual('000100000006f70688b800ff', request.hex())
        request = create_modbus_tcp_request(0xf7, 0x6, 0x88b8, -1)
        self.assertEqual('000100000006f70688b8ffff', request.hex())

    def test_create_modbus_tcp_multi_request(self):
        request = create_modbus_tcp_multi_request(0xf7, 0x10, 0x88b8, b'\x01\x02\x03\x04\x05\x06')
        self.assertEqual('00010000000df71088b8000306010203040506', request.hex())

    def test_validate_modbus_tcp_read_response(self):
        self.assert_tcp_response_ok('000100000007b4030445565345', 0x3, 310, 2)
        self.assert_tcp_response_ok('000100000007b4030400000002', 0x3, 331, 2)
        # technically illegal, but work around Goodwe bug
        self.assert_tcp_response_ok('000100000006f703020000', 0x3, 47510, 1)
        # length too short
        self.assert_tcp_response_partial('000100000007b403040000', 0x03, 331, 2)
        # failure code
        self.assert_tcp_response_rejected('000100000007b4830400000002', 0x03, 331, 2)

    def test_validate_modbus_tcp_write_response(self):
        self.assert_tcp_response_ok('000100000006b40601364556', 0x06, 310, 0x4556)
        self.assert_tcp_response_ok('000100000006f70688b80021', 0x06, 0x88b8, 0x0021)
        #        self.assert_tcp_response_ok('000100000006f70688b800ff', 0x06, 0xb12c, -1)
        # length too short
        self.assert_tcp_response_partial('000100000006f70388b8', 0x6, 0x88b8, 0x0021)
        # wrong value written
        self.assert_rtu_response_fail('aa55f706b12c0012fba4', 0x06, 0xb12c, 0x14)



import asyncio
import logging
import sys
from importlib.metadata import version

from pymodbus.client import AsyncModbusTcpClient

# Force the local files, not pip installed lib
sys.path.insert(0, '..')
sys.path.insert(0, '../../../GoodWe')
import goodwe

logging.basicConfig(
    format="%(asctime)-15s %(funcName)s(%(lineno)d) - %(levelname)s: %(message)s",
    stream=sys.stderr,
    level=getattr(logging, "DEBUG", None),
)
logger = logging.getLogger(__name__)

if sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

try:
    module_ver = version('goodwe')
    print("WARNING !!!")
    print("==============================")
    print(f"You are executing code with installed pip version goodwe:{module_ver}")
    print("You are not testing the local files, if that was what you meant !!!")
    print("==============================")
except ModuleNotFoundError:
    pass


async def pymodbus(ip):
    client = AsyncModbusTcpClient(host=ip)  # Create client object
    await client.connect()  # connect to device, reconnect automatically

    await client.read_holding_registers(35000, 33, slave=0xf7)
    await client.read_holding_registers(47547, 6, slave=0xf7)
    await client.read_holding_registers(47589, 6, slave=0xf7)

    i = 4
    while True:
        logger.info("################################")
        logger.info("          Request %d", i)
        logger.info("################################")
        await client.read_holding_registers(35100, 125, slave=0xf7)
        await client.read_holding_registers(37000, 24, slave=0xf7)
        await client.read_holding_registers(36000, 45, slave=0xf7)
        # await client.read_holding_registers(36000, 58, slave=0xf7)
        # await client.read_holding_registers(35301, 61, slave=0xf7)
        await asyncio.sleep(10)
        i += 1


async def read_modbus_range(ip, port, register, length):
    inverter = await goodwe.connect(host=ip, port=port, family="ET", timeout=1, retries=3)
    # inverter.set_keep_alive(False)

    i = 1
    while True:
        logger.info("################################")
        logger.info("          Request %d", i)
        logger.info("################################")
        await goodwe.protocol.ModbusRtuReadCommand(0xf7, register, length).execute(
            goodwe.protocol.UdpInverterProtocol(ip, port, 1, 3))
        await asyncio.sleep(5)
        i += 1


async def get_runtime_data(ip, port):
    inverter = await goodwe.connect(host=ip, port=port, family="ET", timeout=1, retries=3)
    # inverter.set_keep_alive(False)

    i = 1
    while True:
        logger.info("################################")
        logger.info("          Request %d", i)
        logger.info("################################")
        await inverter.read_runtime_data()
        await asyncio.sleep(5)
        i += 1


# asyncio.run(pymodbus('127.0.0.1'))
# asyncio.run(read_modbus_range('192.168.2.14', 8899, 35100, 125))
asyncio.run(get_runtime_data('127.0.0.1', 502))



from unittest import TestCase

from goodwe.sensor import *


class MockResponse(ProtocolResponse):

    def __init__(self, response: str):
        super().__init__(bytes.fromhex(response), None)

    def response_data(self) -> bytes:
        return self.raw_data


class TestUtils(TestCase):

    def test_byte(self):
        testee = Byte("", 0, "", "", None)

        data = MockResponse("0c")
        self.assertEqual(12, testee.read(data))

        data = MockResponse("f0")
        self.assertEqual(-16, testee.read(data))

    def test_byteH(self):
        testee = ByteH("", 0, "", "", None)

        data = MockResponse("2039")
        self.assertEqual(32, testee.read(data))

        self.assertEqual("2039", testee.encode_value(32, bytes.fromhex("3039")).hex())
        self.assertEqual("2039", testee.encode_value("32", bytes.fromhex("3039")).hex())
        self.assertEqual("ff39", testee.encode_value(-1, bytes.fromhex("3039")).hex())
        self.assertEqual("7f39", testee.encode_value(127, bytes.fromhex("3039")).hex())
        self.assertEqual("20ff", testee.encode_value(32, bytes.fromhex("ffff")).hex())

    def test_byteL(self):
        testee = ByteL("", 0, "", "", None)

        data = MockResponse("307f")
        self.assertEqual(127, testee.read(data))

        self.assertEqual("3020", testee.encode_value(32, bytes.fromhex("3039")).hex())
        self.assertEqual("3020", testee.encode_value("32", bytes.fromhex("3039")).hex())
        self.assertEqual("30ff", testee.encode_value(-1, bytes.fromhex("3039")).hex())
        self.assertEqual("307f", testee.encode_value(127, bytes.fromhex("3039")).hex())
        self.assertEqual("ff20", testee.encode_value(32, bytes.fromhex("ffff")).hex())

    def test_integer(self):
        testee = Integer("", 0, "", "", None)

        data = MockResponse("0031")
        self.assertEqual(49, testee.read(data))
        self.assertEqual("0031", testee.encode_value(49).hex())
        self.assertEqual("0031", testee.encode_value("49").hex())

        data = MockResponse("ff9e")
        self.assertEqual(65438, testee.read(data))
        self.assertEqual("ff9e", testee.encode_value(65438).hex())
        self.assertEqual("ff9e", testee.encode_value("65438").hex())

    def test_integer_signed(self):
        testee = IntegerS("", 0, "", "", None)

        data = MockResponse("0031")
        self.assertEqual(49, testee.read(data))
        self.assertEqual("0031", testee.encode_value(49).hex())
        self.assertEqual("0031", testee.encode_value("49").hex())

        data = MockResponse("ff9e")
        self.assertEqual(-98, testee.read(data))
        self.assertEqual("ff9e", testee.encode_value(-98).hex())
        self.assertEqual("ff9e", testee.encode_value("-98").hex())

    def test_decimal(self):
        testee = Decimal("", 0, 10, "", "", None)

        data = MockResponse("0031")
        self.assertEqual(4.9, testee.read(data))
        self.assertEqual("0031", testee.encode_value(4.9).hex())
        self.assertEqual("0031", testee.encode_value("4.9").hex())

        data = MockResponse("ff9e")
        self.assertEqual(-9.8, testee.read(data))
        self.assertEqual("ff9e", testee.encode_value(-9.8).hex())
        self.assertEqual("ff9e", testee.encode_value("-9.8").hex())

    def test_voltage(self):
        testee = Voltage("", 0, "", None)

        data = MockResponse("0cfe")
        self.assertEqual(332.6, testee.read(data))
        self.assertEqual("0cfe", testee.encode_value(332.6).hex())
        self.assertEqual("0cfe", testee.encode_value("332.6").hex())

        data = MockResponse("1f64")
        self.assertEqual(803.6, testee.read(data))
        self.assertEqual("1f64", testee.encode_value(803.6).hex())
        self.assertEqual("1f64", testee.encode_value("803.6").hex())

        data = MockResponse("a000")
        self.assertEqual(4096.0, testee.read(data))

        data = MockResponse("ffff")
        self.assertEqual(0, testee.read(data))

    def test_current(self):
        testee = Current("", 0, "", None)

        data = MockResponse("0031")
        self.assertEqual(4.9, testee.read(data))
        self.assertEqual("0031", testee.encode_value(4.9).hex())
        self.assertEqual("0031", testee.encode_value("4.9").hex())

        data = MockResponse("ff9e")
        self.assertEqual(6543.8, testee.read(data))
        self.assertEqual("ff9e", testee.encode_value(6543.8).hex())
        self.assertEqual("ff9e", testee.encode_value("6543.8").hex())

        data = MockResponse("ffff")
        self.assertEqual(0, testee.read(data))

    def test_current_signed(self):
        testee = CurrentS("", 0, "", None)

        data = MockResponse("0031")
        self.assertEqual(4.9, testee.read(data))
        self.assertEqual("0031", testee.encode_value(4.9).hex())
        self.assertEqual("0031", testee.encode_value("4.9").hex())

        data = MockResponse("ff9e")
        self.assertEqual(-9.8, testee.read(data))
        self.assertEqual("ff9e", testee.encode_value(-9.8).hex())
        self.assertEqual("ff9e", testee.encode_value("-9.8").hex())

    def test_power4(self):
        testee = Power4("", 0, "", None)

        data = MockResponse("0000069f")
        self.assertEqual(1695, testee.read(data))

        data = MockResponse("fffffffd")
        self.assertEqual(4294967293, testee.read(data))

        data = MockResponse("ffffffff")
        self.assertIsNone(testee.read(data))

    def test_power4_signed(self):
        testee = Power4S("", 0, "", None)

        data = MockResponse("0000069f")
        self.assertEqual(1695, testee.read(data))

        data = MockResponse("fffffffd")
        self.assertEqual(-3, testee.read(data))

    def test_energy(self):
        testee = Energy("", 0, "", None)

        data = MockResponse("0972")
        self.assertEqual(241.8, testee.read(data))

    def test_energy4(self):
        testee = Energy4("", 0, "", None)

        data = MockResponse("00020972")
        self.assertEqual(13349.0, testee.read(data))
        data = MockResponse("ffffffff")
        self.assertIsNone(testee.read(data))

    def test_energy8(self):
        testee = Energy8("", 0, "", None)

        data = MockResponse("0000000000015b41")
        self.assertEqual(888.97, testee.read(data))
        data = MockResponse("0000000000038E6C")
        self.assertEqual(2330.68, testee.read(data))
        data = MockResponse("ffffffffffffffff")
        self.assertIsNone(testee.read(data))

    def test_temp(self):
        testee = Temp("", 0, "", None)

        data = MockResponse("0177")
        self.assertEqual(37.5, testee.read(data))
        data = MockResponse("ffff")
        self.assertIsNone(testee.read(data))

    def test_timestamp(self):
        testee = Timestamp("", 0, "", None)

        data = MockResponse("160104121e19")
        self.assertEqual(datetime(2022, 1, 4, 18, 30, 25), testee.read(data))
        self.assertEqual("160104121e19", testee.encode_value(datetime(2022, 1, 4, 18, 30, 25)).hex())
        self.assertEqual("160104121e19", testee.encode_value("2022-01-04T18:30:25").hex())

    def test_eco_mode_v1(self):
        testee = EcoModeV1("", 0, "")

        data = MockResponse("0d1e0e28ffc4ff1a")
        self.assertEqual("13:30-14:40 Mon,Wed,Thu -60% On", testee.read(data).__str__())
        self.assertEqual(bytes.fromhex("0d1e0e28ffc4ff1a"), testee.encode_value(bytes.fromhex("0d1e0e28ffc4ff1a")))
        self.assertRaises(ValueError, lambda: testee.encode_value("some string"))
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())

        data = MockResponse(testee.encode_charge(-40).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% On", testee.read(data).__str__())
        self.assertTrue(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% (SoC 100%) On",
                         testee.as_eco_mode_v2().__str__())
        data = MockResponse(testee.encode_discharge(60).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat 60% On", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertTrue(testee.read(data).is_eco_discharge_mode())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat 60% (SoC 100%) On",
                         testee.as_eco_mode_v2().__str__())
        data = MockResponse(testee.encode_off().hex())
        self.assertEqual("48:0-48:0  100% Off", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())

    def test_schedule(self):
        testee = Schedule("", 0, "")

        data = MockResponse("0d1e0e28ff1affc4005a0000")
        self.assertEqual("13:30-14:40 Mon,Wed,Thu -60% (SoC 90%) On", testee.read(data).__str__())
        self.assertEqual(ScheduleType.ECO_MODE, testee.schedule_type)
        self.assertEqual(bytes.fromhex("0d1e0e28ff1affc4005a0000"),
                         testee.encode_value(bytes.fromhex("0d1e0e28ff1affc4005a0000")))
        self.assertRaises(ValueError, lambda: testee.encode_value("some string"))
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())

        data = MockResponse(testee.encode_charge(-40, 80).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% (SoC 80%) On", testee.read(data).__str__())
        self.assertTrue(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% On", testee.as_eco_mode_v1().__str__())
        data = MockResponse(testee.encode_discharge(60).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat 60% (SoC 100%) On", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertTrue(testee.read(data).is_eco_discharge_mode())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat 60% On", testee.as_eco_mode_v1().__str__())
        data = MockResponse(testee.encode_off().hex())
        self.assertEqual("48:0-48:0  100% (SoC 100%) Off", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())

        data = MockResponse("0300080006fefd12005fcfff")
        self.assertEqual("3:0-8:0 Mon -75% (SoC 95%) Off", testee.read(data).__str__())
        data = MockResponse("0000173b5500001400640000")
        self.assertEqual("0:0-23:59  20% (SoC 100%) Unset", testee.read(data).__str__())
        data = MockResponse("ffffffff557f000000010001")
        self.assertEqual("-1:-1--1:-1 Sun,Mon,Tue,Wed,Thu,Fri,Sat Jan 0% (SoC 1%) Unset", testee.read(data).__str__())
        data = MockResponse("000000005500000000000000")
        self.assertEqual("0:0-0:0  0% (SoC 0%) Unset", testee.read(data).__str__())

    def test_eco_mode_v745(self):
        testee = Schedule("", 0, "", ScheduleType.ECO_MODE_745)

        data = MockResponse("0d1e0e28f91affc4005a0000")
        self.assertEqual("13:30-14:40 Mon,Wed,Thu -6% (SoC 90%) On", testee.read(data).__str__())
        self.assertEqual(bytes.fromhex("0d1e0e28f91affc4005a0000"),
                         testee.encode_value(bytes.fromhex("0d1e0e28f91affc4005a0000")))
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)

        data = MockResponse(testee.encode_charge(-40, 80).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% (SoC 80%) On", testee.read(data).__str__())
        self.assertTrue(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse(testee.encode_discharge(60).hex())
        self.assertEqual("0:0-23:59 Sun,Mon,Tue,Wed,Thu,Fri,Sat 60% (SoC 100%) On", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertTrue(testee.read(data).is_eco_discharge_mode())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse(testee.encode_off().hex())
        self.assertEqual("48:0-48:0  100% (SoC 100%) Off", testee.read(data).__str__())
        self.assertFalse(testee.read(data).is_eco_charge_mode())
        self.assertFalse(testee.read(data).is_eco_discharge_mode())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        self.assertEqual(1000, testee.power)
        self.assertEqual(100, testee.get_power())
        self.assertEqual("%", testee.get_power_unit())

        data = MockResponse("10001600f97f00c800000fff")
        self.assertEqual("16:0-22:0 Sun,Mon,Tue,Wed,Thu,Fri,Sat 20% (SoC 0%) On", testee.read(data).__str__())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse("10001600067f00c800000fff")
        self.assertEqual("16:0-22:0 Sun,Mon,Tue,Wed,Thu,Fri,Sat 20% (SoC 0%) Off", testee.read(data).__str__())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse("10001600f97ffe70004b0fff")
        self.assertEqual("16:0-22:0 Sun,Mon,Tue,Wed,Thu,Fri,Sat -40% (SoC 75%) On", testee.read(data).__str__())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse("10001600f97fff3800320fff")
        self.assertEqual("16:0-22:0 Sun,Mon,Tue,Wed,Thu,Fri,Sat -20% (SoC 50%) On", testee.read(data).__str__())
        self.assertEqual(ScheduleType.ECO_MODE_745, testee.schedule_type)
        data = MockResponse("10001600f902ff38004b0002")
        self.assertEqual("16:0-22:0 Mon Feb -20% (SoC 75%) On", testee.read(data).__str__())
        data = MockResponse("10001600f902ff38004b0004")
        self.assertEqual("16:0-22:0 Mon Mar -20% (SoC 75%) On", testee.read(data).__str__())

    def test_peak_shaving_mode(self):
        testee = PeakShavingMode("", 0, "")

        data = MockResponse("010a020a037f00fa00370000")
        self.assertEqual("1:10-2:10 Sun,Mon,Tue,Wed,Thu,Fri,Sat 2500W (SoC 55%) Off", testee.read(data).__str__())
        self.assertEqual(bytes.fromhex("010a020a037f00fa00370000"),
                         testee.encode_value(bytes.fromhex("010a020a037f00fa00370000")))

        data = MockResponse("00000d08fc7f006400140000")
        self.assertEqual("0:0-13:8 Sun,Mon,Tue,Wed,Thu,Fri,Sat 1000W (SoC 20%) On", testee.read(data).__str__())
        self.assertEqual(ScheduleType.PEAK_SHAVING, testee.schedule_type)
        data = MockResponse("00000d08037f000000000000")
        self.assertEqual("0:0-13:8 Sun,Mon,Tue,Wed,Thu,Fri,Sat 0W (SoC 0%) Off", testee.read(data).__str__())
        self.assertEqual(ScheduleType.PEAK_SHAVING, testee.schedule_type)

    def test_decode_bitmap(self):
        self.assertEqual('', decode_bitmap(0, ERROR_CODES))
        self.assertEqual('Utility Loss', decode_bitmap(512, ERROR_CODES))
        self.assertEqual('Utility Loss', decode_bitmap(516, ERROR_CODES))
        self.assertEqual('Utility Loss, Vac Failure', decode_bitmap(131584, ERROR_CODES))
        self.assertEqual('err16', decode_bitmap(65536, BMS_WARNING_CODES))



"""Simple test script to scan inverter present on local network"""
import asyncio
import logging
import sys

import goodwe
from goodwe.exceptions import InverterError
from goodwe.protocol import ProtocolCommand, UdpInverterProtocol

logging.basicConfig(
    format="%(asctime)-15s %(funcName)s(%(lineno)d) - %(levelname)s: %(message)s",
    stream=sys.stderr,
    level=getattr(logging, "DEBUG", None),
)

if sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


def try_command(command, ip):
    print(f"Trying command: {command}")
    try:
        response = asyncio.run(
            ProtocolCommand(bytes.fromhex(command), lambda x: True).execute(UdpInverterProtocol(ip, 8899)))
        print(f"Response to {command} command: {response.raw_data.hex()}")
    except InverterError:
        print(f"No response to {command} command")


result = asyncio.run(goodwe.search_inverters()).decode("utf-8").split(",")
print(f"Located inverter at IP: {result[0]}, mac: {result[1]}, name: {result[2]}")

# EM/ES
try_command("AA55C07F0102000241", result[0])
# DT (SolarGo)
try_command("7F03753100280409", result[0])

print(f"Identifying inverter at IP: {result[0]}")
inverter = asyncio.run(goodwe.discover(result[0]))
print(
    f"Identified inverter model: {inverter.model_name}, serialNr: {inverter.serial_number}"
)



from __future__ import annotations

import logging
from typing import Tuple

from .exceptions import InverterError, RequestFailedException, RequestRejectedException
from .inverter import Inverter
from .inverter import OperationMode
from .inverter import SensorKind as Kind
from .modbus import ILLEGAL_DATA_ADDRESS
from .model import is_3_mppt, is_single_phase
from .protocol import ProtocolCommand
from .sensor import *

logger = logging.getLogger(__name__)


class DT(Inverter):
    """Class representing inverter of DT/MS/D-NS/XS or GE's GEP(PSB/PSC) families"""

    __all_sensors: Tuple[Sensor, ...] = (
        Timestamp("timestamp", 30100, "Timestamp"),
        Voltage("vpv1", 30103, "PV1 Voltage", Kind.PV),
        Current("ipv1", 30104, "PV1 Current", Kind.PV),
        Calculated("ppv1",
                   lambda data: round(read_voltage(data, 30103) * read_current(data, 30104)),
                   "PV1 Power", "W", Kind.PV),
        Voltage("vpv2", 30105, "PV2 Voltage", Kind.PV),
        Current("ipv2", 30106, "PV2 Current", Kind.PV),
        Calculated("ppv2",
                   lambda data: round(read_voltage(data, 30105) * read_current(data, 30106)),
                   "PV2 Power", "W", Kind.PV),
        Voltage("vpv3", 30107, "PV3 Voltage", Kind.PV),
        Current("ipv3", 30108, "PV3 Current", Kind.PV),
        Calculated("ppv3",
                   lambda data: round(read_voltage(data, 30107) * read_current(data, 30108)),
                   "PV3 Power", "W", Kind.PV),
        # ppv1 + ppv2 + ppv3
        Calculated("ppv",
                   lambda data: (round(read_voltage(data, 30103) * read_current(data, 30104))) + (round(
                       read_voltage(data, 30105) * read_current(data, 30106))) + (round(
                       read_voltage(data, 30107) * read_current(data, 30108))),
                   "PV Power", "W", Kind.PV),
        # Voltage("vpv4", 14, "PV4 Voltage", Kind.PV),
        # Current("ipv4", 16, "PV4 Current", Kind.PV),
        # Voltage("vpv5", 14, "PV5 Voltage", Kind.PV),
        # Current("ipv5", 16, "PV5 Current", Kind.PV),
        # Voltage("vpv6", 14, "PV6 Voltage", Kind.PV),
        # Current("ipv6", 16, "PV7 Current", Kind.PV),
        Voltage("vline1", 30115, "On-grid L1-L2 Voltage", Kind.AC),
        Voltage("vline2", 30116, "On-grid L2-L3 Voltage", Kind.AC),
        Voltage("vline3", 30117, "On-grid L3-L1 Voltage", Kind.AC),
        Voltage("vgrid1", 30118, "On-grid L1 Voltage", Kind.AC),
        Voltage("vgrid2", 30119, "On-grid L2 Voltage", Kind.AC),
        Voltage("vgrid3", 30120, "On-grid L3 Voltage", Kind.AC),
        Current("igrid1", 30121, "On-grid L1 Current", Kind.AC),
        Current("igrid2", 30122, "On-grid L2 Current", Kind.AC),
        Current("igrid3", 30123, "On-grid L3 Current", Kind.AC),
        Frequency("fgrid1", 30124, "On-grid L1 Frequency", Kind.AC),
        Frequency("fgrid2", 30125, "On-grid L2 Frequency", Kind.AC),
        Frequency("fgrid3", 30126, "On-grid L3 Frequency", Kind.AC),
        Calculated("pgrid1",
                   lambda data: round(read_voltage(data, 30118) * read_current(data, 30121)),
                   "On-grid L1 Power", "W", Kind.AC),
        Calculated("pgrid2",
                   lambda data: round(read_voltage(data, 30119) * read_current(data, 30122)),
                   "On-grid L2 Power", "W", Kind.AC),
        Calculated("pgrid3",
                   lambda data: round(read_voltage(data, 30120) * read_current(data, 30123)),
                   "On-grid L3 Power", "W", Kind.AC),
        # 30127 reserved
        PowerS("total_inverter_power", 30128, "Total Power", Kind.AC),
        Integer("work_mode", 30129, "Work Mode code"),
        Enum2("work_mode_label", 30129, WORK_MODES, "Work Mode"),
        Long("error_codes", 30130, "Error Codes"),
        Integer("warning_code", 30132, "Warning code"),
        Apparent4("apparent_power", 30133, "Apparent Power", Kind.AC),
        Reactive4("reactive_power", 30135, "Reactive Power", Kind.AC),
        # 30137 reserved
        # 30138 reserved
        Decimal("power_factor", 30139, 1000, "Power Factor", "", Kind.GRID),
        # 30140 reserved
        Temp("temperature", 30141, "Inverter Temperature", Kind.AC),
        # 30142 reserved
        # 30143 reserved
        Energy("e_day", 30144, "Today's PV Generation", Kind.PV),
        Energy4("e_total", 30145, "Total PV Generation", Kind.PV),
        Long("h_total", 30147, "Hours Total", "h", Kind.PV),
        Integer("safety_country", 30149, "Safety Country code", "", Kind.AC),
        Enum2("safety_country_label", 30149, SAFETY_COUNTRIES, "Safety Country", Kind.AC),
        # 30150 reserved
        # 30151 reserved
        # 30152 reserved
        # 30153 reserved
        # 30154 reserved
        # 30155 reserved
        # 30156 reserved
        # 30157 reserved
        # 30158 reserved
        # 30159 reserved
        # 30160 reserved
        # 30161 reserved
        Integer("funbit", 30162, "FunBit", "", Kind.PV),
        Voltage("vbus", 30163, "Bus Voltage", Kind.PV),
        Voltage("vnbus", 30164, "NBus Voltage", Kind.PV),
        Long("derating_mode", 30165, "Derating Mode code"),
        EnumBitmap4("derating_mode_label", 30165, DERATING_MODE_CODES, "Derating Mode"),
        # 30167 reserved
        # 30168 reserved
        # 30169 reserved
        # 30170 reserved
        # 30171 reserved
        # 30172 reserved
    )

    # Inverter's meter data
    # Modbus registers from offset 0x75f4 (30196)
    __all_sensors_meter: Tuple[Sensor, ...] = (
        PowerS("active_power", 30196, "Active Power", Kind.GRID),
    )

    # Modbus registers of inverter settings, offsets are modbus register addresses
    __all_settings: Tuple[Sensor, ...] = (
        Timestamp("time", 40313, "Inverter time"),

        Integer("shadow_scan", 40326, "Shadow Scan", "", Kind.PV),
        Integer("grid_export", 40327, "Grid Export Enabled", "", Kind.GRID),
        Integer("grid_export_limit", 40328, "Grid Export Limit", "%", Kind.GRID),
        Integer("start", 40330, "Start / Power On", "", Kind.GRID),
        Integer("stop", 40331, "Stop / Power Off", "", Kind.GRID),
        Integer("restart", 40332, "Restart", "", Kind.GRID),
        Integer("grid_export_hw", 40345, "Grid Export Enabled (HW)", "", Kind.GRID),
    )

    # Settings for single phase inverters
    __settings_single_phase: Tuple[Sensor, ...] = (
        Long("grid_export_limit", 40328, "Grid Export Limit", "W", Kind.GRID),
    )

    # Settings for three phase inverters
    __settings_three_phase: Tuple[Sensor, ...] = (
        Integer("grid_export_limit", 40336, "Grid Export Limit", "%", Kind.GRID),
    )

    def __init__(self, host: str, port: int, comm_addr: int = 0, timeout: int = 1, retries: int = 3):
        super().__init__(host, port, comm_addr if comm_addr else 0x7f, timeout, retries)
        self._READ_DEVICE_VERSION_INFO: ProtocolCommand = self._read_command(0x7531, 0x0028)
        self._READ_RUNNING_DATA: ProtocolCommand = self._read_command(0x7594, 0x0049)
        self._READ_METER_DATA: ProtocolCommand = self._read_command(0x75f4, 0x01)
        self._sensors = self.__all_sensors
        self._sensors_meter = self.__all_sensors_meter
        self._settings: dict[str, Sensor] = {s.id_: s for s in self.__all_settings}
        self._has_meter: bool = True

    @staticmethod
    def _single_phase_only(s: Sensor) -> bool:
        """Filter to exclude phase2/3 sensors on single phase inverters"""
        return not ((s.id_.endswith('2') or s.id_.endswith('3')) and 'pv' not in s.id_)

    @staticmethod
    def _pv1_pv2_only(s: Sensor) -> bool:
        """Filter to exclude sensors on < 3 PV inverters"""
        return not s.id_.endswith('pv3')

    async def read_device_info(self):
        response = await self._read_from_socket(self._READ_DEVICE_VERSION_INFO)
        response = response.response_data()
        try:
            self.model_name = response[22:32].decode("ascii").rstrip()
        except:
            print("No model name sent from the inverter.")
        # Modbus registers from 30001 - 30040
        self.serial_number = self._decode(response[6:22])  # 30004 - 30012
        self.dsp1_version = read_unsigned_int(response, 66)  # 30034
        self.dsp2_version = read_unsigned_int(response, 68)  # 30035
        self.arm_version = read_unsigned_int(response, 70)  # 30036
        self.dsp_svn_version = read_unsigned_int(response, 72)  # 35037
        self.arm_svn_version = read_unsigned_int(response, 74)  # 35038
        self.firmware = "{}.{}.{:02x}".format(self.dsp1_version, self.dsp2_version, self.arm_version)

        if is_single_phase(self):
            # this is single phase inverter, filter out all L2 and L3 sensors
            self._sensors = tuple(filter(self._single_phase_only, self.__all_sensors))
            self._settings.update({s.id_: s for s in self.__settings_single_phase})
        else:
            self._settings.update({s.id_: s for s in self.__settings_three_phase})

        if is_3_mppt(self):
            # this is 3 PV strings inverter, keep all sensors
            pass
        else:
            # this is only 2 PV strings inverter
            self._sensors = tuple(filter(self._pv1_pv2_only, self._sensors))
        pass

    async def read_runtime_data(self) -> Dict[str, Any]:
        response = await self._read_from_socket(self._READ_RUNNING_DATA)
        data = self._map_response(response, self._sensors)

        if self._has_meter:
            try:
                response = await self._read_from_socket(self._READ_METER_DATA)
                data.update(self._map_response(response, self._sensors_meter))
            except (RequestRejectedException, RequestFailedException):
                logger.info("Meter values not supported, disabling further attempts.")
                self._has_meter = False

        return data

    async def read_setting(self, setting_id: str) -> Any:
        setting = self._settings.get(setting_id)
        if setting:
            return await self._read_setting(setting)
        else:
            if setting_id.startswith("modbus"):
                response = await self._read_from_socket(self._read_command(int(setting_id[7:]), 1))
                return int.from_bytes(response.read(2), byteorder="big", signed=True)
            else:
                raise ValueError(f'Unknown setting "{setting_id}"')

    async def _read_setting(self, setting: Sensor) -> Any:
        try:
            count = (setting.size_ + (setting.size_ % 2)) // 2
            response = await self._read_from_socket(self._read_command(setting.offset, count))
            return setting.read_value(response)
        except RequestRejectedException as ex:
            if ex.message == ILLEGAL_DATA_ADDRESS:
                logger.debug("Unsupported setting %s", setting.id_)
                self._settings.pop(setting.id_, None)
                raise ValueError(f'Unknown setting "{setting.id_}"')
            return None

    async def write_setting(self, setting_id: str, value: Any):
        setting = self._settings.get(setting_id)
        if setting:
            await self._write_setting(setting, value)
        else:
            if setting_id.startswith("modbus"):
                await self._read_from_socket(self._write_command(int(setting_id[7:]), int(value)))
            else:
                raise ValueError(f'Unknown setting "{setting_id}"')

    async def _write_setting(self, setting: Sensor, value: Any):
        if setting.size_ == 1:
            # modbus can address/store only 16 bit values, read the other 8 bytes
            response = await self._read_from_socket(self._read_command(setting.offset, 1))
            raw_value = setting.encode_value(value, response.response_data()[0:2])
        else:
            raw_value = setting.encode_value(value)
        if len(raw_value) <= 2:
            value = int.from_bytes(raw_value, byteorder="big", signed=True)
            await self._read_from_socket(self._write_command(setting.offset, value))
        else:
            await self._read_from_socket(self._write_multi_command(setting.offset, raw_value))

    async def read_settings_data(self) -> Dict[str, Any]:
        data = {}
        for setting in self.settings():
            value = await self.read_setting(setting.id_)
            data[setting.id_] = value
        return data

    async def get_grid_export_limit(self) -> int:
        return await self.read_setting('grid_export_limit')

    async def set_grid_export_limit(self, export_limit: int) -> None:
        if export_limit >= 0:
            return await self.write_setting('grid_export_limit', export_limit)

    async def get_operation_modes(self, include_emulated: bool) -> Tuple[OperationMode, ...]:
        return ()

    async def get_operation_mode(self) -> OperationMode:
        raise InverterError("Operation not supported.")

    async def set_operation_mode(self, operation_mode: OperationMode, eco_mode_power: int = 100,
                                 eco_mode_soc: int = 100) -> None:
        raise InverterError("Operation not supported.")

    async def get_ongrid_battery_dod(self) -> int:
        raise InverterError("Operation not supported, inverter has no batteries.")

    async def set_ongrid_battery_dod(self, dod: int) -> None:
        raise InverterError("Operation not supported, inverter has no batteries.")

    def sensors(self) -> Tuple[Sensor, ...]:
        result = self._sensors
        if self._has_meter:
            result = result + self._sensors_meter
        return result

    def settings(self) -> Tuple[Sensor, ...]:
        return tuple(self._settings.values())



# Serial number tags to identify inverter type
from .inverter import Inverter

PLATFORM_105_MODELS = ("ESU", "EMU", "ESA", "BPS", "BPU", "EMJ", "IJL")
PLATFORM_205_MODELS = ("ETU", "ETL", "ETR", "BHN", "EHU", "BHU", "EHR", "BTU")
PLATFORM_745_LV_MODELS = ("ESN", "EBN", "EMN", "SPN", "ERN", "ESC", "HLB", "HMB", "HBB", "EOA")
PLATFORM_745_HV_MODELS = ("ETT", "HTA", "HUB", "AEB", "SPB", "CUB", "EUB", "HEB", "ERB", "BTT", "ETF", "ARB", "URB",
                          "EBR")
PLATFORM_753_MODELS = ("AES", "HHI", "ABP", "EHB", "HSB", "HUA", "CUA")

ET_MODEL_TAGS = PLATFORM_205_MODELS + PLATFORM_745_LV_MODELS + PLATFORM_745_HV_MODELS + PLATFORM_753_MODELS + (
    "ETC", "BTC", "BTN")  # Qianhai
ES_MODEL_TAGS = PLATFORM_105_MODELS
DT_MODEL_TAGS = ("DTU", "DTS",
                 "MSU", "MST", "MSC", "DSN", "DTN", "DST", "NSU", "SSN", "SST", "SSX", "SSY",
                 "PSB", "PSC")

SINGLE_PHASE_MODELS = ("DSN", "DST", "NSU", "SSN", "SST", "SSX", "SSY",  # DT
                       "MSU", "MST", "PSB", "PSC",
                       "MSC",  # Found on third gen MS
                       "EHU", "EHR", "HSB",  # ET
                       "ESN", "EMN", "ERN", "EBN", "HLB", "HMB", "HBB", "SPN")  # ES Gen 2

MPPT3_MODELS = ("MSU", "MST", "PSC", "MSC",
                "25KET", "29K9ET")

MPPT4_MODELS = ("HSB",)

BAT_2_MODELS = ("25KET", "29K9ET")


def is_single_phase(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in SINGLE_PHASE_MODELS)


def is_3_mppt(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in MPPT3_MODELS)


def is_4_mppt(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in MPPT4_MODELS)


def is_2_battery(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in BAT_2_MODELS)


def is_745_platform(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in PLATFORM_745_LV_MODELS) or any(
        model in inverter.serial_number for model in PLATFORM_745_HV_MODELS)


def is_753_platform(inverter: Inverter) -> bool:
    return any(model in inverter.serial_number for model in PLATFORM_753_MODELS)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import Any, Callable, Dict, Tuple, Optional

from .exceptions import MaxRetriesException, RequestFailedException
from .protocol import InverterProtocol, ProtocolCommand, ProtocolResponse, TcpInverterProtocol, UdpInverterProtocol

logger = logging.getLogger(__name__)


class SensorKind(Enum):
    """
    Enumeration of sensor kinds.

    Possible values are:
    PV - inverter photo-voltaic (e.g. dc voltage of pv panels)
    AC - inverter grid output (e.g. ac voltage of grid connected output)
    UPS - inverter ups/eps/backup output (e.g. ac voltage of backup/off-grid connected output)
    BAT - battery (e.g. dc voltage of connected battery pack)
    GRID - power grid/smart meter (e.g. active power exported to grid)
    BMS - BMS direct data (e.g. dc voltage of)
    """

    PV = 1
    AC = 2
    UPS = 3
    BAT = 4
    GRID = 5
    BMS = 6


@dataclass
class Sensor:
    """Definition of inverter sensor and its attributes"""

    id_: str
    offset: int
    name: str
    size_: int
    unit: str
    kind: Optional[SensorKind]

    def read_value(self, data: ProtocolResponse) -> Any:
        """Read the sensor value from data at current position"""
        raise NotImplementedError()

    def read(self, data: ProtocolResponse) -> Any:
        """Read the sensor value from data (at sensor offset)"""
        data.seek(self.offset)
        return self.read_value(data)

    def encode_value(self, value: Any, register_value: bytes = None) -> bytes:
        """Encode the (setting mostly) value to (usually) 2 bytes raw register value"""
        raise NotImplementedError()


class OperationMode(IntEnum):
    """
    Enumeration of sensor kinds.

    Possible values are:
    GENERAL - General mode
    OFF_GRID - Off grid mode
    BACKUP - Backup mode
    ECO - Eco mode
    PEAK_SHAVING - Peak shaving mode
    ECO_CHARGE - Eco mode with a single "Charge" group valid all the time (from 00:00-23:59, Mon-Sun)
    ECO_DISCHARGE - Eco mode with a single "Discharge" group valid all the time (from 00:00-23:59, Mon-Sun)
    """

    GENERAL = 0
    OFF_GRID = 1
    BACKUP = 2
    ECO = 3
    PEAK_SHAVING = 4
    SELF_USE = 5
    ECO_CHARGE = 98
    ECO_DISCHARGE = 99


class Inverter(ABC):
    """
    Common superclass for various inverter models implementations.
    Represents the inverter state and its basic behavior
    """

    def __init__(self, host: str, port: int, comm_addr: int = 0, timeout: int = 1, retries: int = 3):
        self._protocol: InverterProtocol = self._create_protocol(host, port, comm_addr, timeout, retries)
        self._consecutive_failures_count: int = 0

        self.model_name: str | None = None
        self.serial_number: str | None = None
        self.rated_power: int = 0
        self.ac_output_type: int | None = None
        self.firmware: str | None = None
        self.arm_firmware: str | None = None
        self.modbus_version: int | None = None
        self.dsp1_version: int = 0
        self.dsp2_version: int = 0
        self.dsp_svn_version: int | None = None
        self.arm_version: int = 0
        self.arm_svn_version: int | None = None

    def _read_command(self, offset: int, count: int) -> ProtocolCommand:
        """Create read protocol command."""
        return self._protocol.read_command(offset, count)

    def _write_command(self, register: int, value: int) -> ProtocolCommand:
        """Create write protocol command."""
        return self._protocol.write_command(register, value)

    def _write_multi_command(self, offset: int, values: bytes) -> ProtocolCommand:
        """Create write multiple protocol command."""
        return self._protocol.write_multi_command(offset, values)

    async def _read_from_socket(self, command: ProtocolCommand) -> ProtocolResponse:
        try:
            result = await command.execute(self._protocol)
            self._consecutive_failures_count = 0
            return result
        except MaxRetriesException:
            self._consecutive_failures_count += 1
            raise RequestFailedException(f'No valid response received even after {self._protocol.retries} retries',
                                         self._consecutive_failures_count) from None
        except RequestFailedException as ex:
            self._consecutive_failures_count += 1
            raise RequestFailedException(ex.message, self._consecutive_failures_count) from None

    def set_keep_alive(self, keep_alive: bool) -> None:
        self._protocol.keep_alive = keep_alive

    @abstractmethod
    async def read_device_info(self):
        """
        Request the device information from the inverter.
        The inverter instance variables will be loaded with relevant data.
        """
        raise NotImplementedError()

    @abstractmethod
    async def read_runtime_data(self) -> Dict[str, Any]:
        """
        Request the runtime data from the inverter.
        Answer dictionary of individual sensors and their values.
        List of supported sensors (and their definitions) is provided by sensors() method.
        """
        raise NotImplementedError()

    @abstractmethod
    async def read_setting(self, setting_id: str) -> Any:
        """
        Read the value of specific inverter setting/configuration parameter.
        Setting must be in list provided by settings() method, otherwise ValueError is raised.
        """
        raise NotImplementedError()

    @abstractmethod
    async def write_setting(self, setting_id: str, value: Any):
        """
        Set the value of specific inverter settings/configuration parameter.
        Setting must be in list provided by settings() method, otherwise ValueError is raised.

        BEWARE !!!
        This method modifies inverter operational parameter (usually accessible to installers only).
        Use with caution and at your own risk !
        """
        raise NotImplementedError()

    @abstractmethod
    async def read_settings_data(self) -> Dict[str, Any]:
        """
        Request the settings data from the inverter.
        Answer dictionary of individual settings and their values.
        List of supported settings (and their definitions) is provided by settings() method.
        """
        raise NotImplementedError()

    async def send_command(
            self, command: bytes, validator: Callable[[bytes], bool] = lambda x: True
    ) -> ProtocolResponse:
        """
        Send low level command (as bytes).
        Answer ProtocolResponse with command's raw response data.
        """
        return await self._read_from_socket(ProtocolCommand(command, validator))

    @abstractmethod
    async def get_grid_export_limit(self) -> int:
        """
        Get the current grid export limit in W
        """
        raise NotImplementedError()

    @abstractmethod
    async def set_grid_export_limit(self, export_limit: int) -> None:
        """
        BEWARE !!!
        This method modifies inverter operational parameter accessible to installers only.
        Use with caution and at your own risk !

        Set the grid export limit in W
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_operation_modes(self, include_emulated: bool) -> Tuple[OperationMode, ...]:
        """
        Answer list of supported inverter operation modes
        """
        return ()

    @abstractmethod
    async def get_operation_mode(self) -> OperationMode:
        """
        Get the inverter operation mode
        """
        raise NotImplementedError()

    @abstractmethod
    async def set_operation_mode(self, operation_mode: OperationMode, eco_mode_power: int = 100,
                                 eco_mode_soc: int = 100) -> None:
        """
        BEWARE !!!
        This method modifies inverter operational parameter accessible to installers only.
        Use with caution and at your own risk !

        Set the inverter operation mode

        The modes ECO_CHARGE and ECO_DISCHARGE are not real inverter operation modes, but a convenience
        shortcuts to enter Eco Mode with a single group valid all the time (from 00:00-23:59, Mon-Sun)
        charging or discharging with optional charging power and SoC (%) parameters.
        """
        raise NotImplementedError()

    @abstractmethod
    async def get_ongrid_battery_dod(self) -> int:
        """
        Get the On-Grid Battery DoD
        0% - 89%
        """
        raise NotImplementedError()

    @abstractmethod
    async def set_ongrid_battery_dod(self, dod: int) -> None:
        """
        BEWARE !!!
        This method modifies On-Grid Battery DoD parameter accessible to installers only.
        Use with caution and at your own risk !

        Set the On-Grid Battery DoD
        0% - 89%
        """
        raise NotImplementedError()

    @abstractmethod
    def sensors(self) -> Tuple[Sensor, ...]:
        """
        Return tuple of sensor definitions
        """
        raise NotImplementedError()

    @abstractmethod
    def settings(self) -> Tuple[Sensor, ...]:
        """
        Return tuple of settings definitions
        """
        raise NotImplementedError()

    @staticmethod
    def _create_protocol(host: str, port: int, comm_addr: int, timeout: int, retries: int) -> InverterProtocol:
        if port == 502:
            return TcpInverterProtocol(host, port, comm_addr, timeout, retries)
        else:
            return UdpInverterProtocol(host, port, comm_addr, timeout, retries)

    @staticmethod
    def _map_response(response: ProtocolResponse, sensors: Tuple[Sensor, ...]) -> Dict[str, Any]:
        """Process the response data and return dictionary with runtime values"""
        result = {}
        for sensor in sensors:
            try:
                result[sensor.id_] = sensor.read(response)
            except ValueError:
                logger.exception("Error reading sensor %s.", sensor.id_)
                result[sensor.id_] = None
        return result

    @staticmethod
    def _decode(data: bytes) -> str:
        """Decode the bytes to ascii string"""
        try:
            if any(x < 32 for x in data):
                return data.decode("utf-16be").rstrip().replace('\x00', '')
            return data.decode("ascii").rstrip()
        except ValueError:
            return data.hex()



from __future__ import annotations

import asyncio
import io
import logging
import platform
import socket
from asyncio.futures import Future
from typing import Tuple, Optional, Callable

from .exceptions import MaxRetriesException, PartialResponseException, RequestFailedException, RequestRejectedException
from .modbus import create_modbus_rtu_request, create_modbus_rtu_multi_request, create_modbus_tcp_request, \
    create_modbus_tcp_multi_request, validate_modbus_rtu_response, validate_modbus_tcp_response, MODBUS_READ_CMD, \
    MODBUS_WRITE_CMD, MODBUS_WRITE_MULTI_CMD

logger = logging.getLogger(__name__)

_modbus_tcp_tx = 0


def _next_tx() -> bytes:
    global _modbus_tcp_tx
    _modbus_tcp_tx += 1
    if _modbus_tcp_tx == 0xFFFF:
        _modbus_tcp_tx = 1
    return int.to_bytes(_modbus_tcp_tx, length=2, byteorder="big", signed=False)


class InverterProtocol:

    def __init__(self, host: str, port: int, comm_addr: int, timeout: int, retries: int):
        self._host: str = host
        self._port: int = port
        self._comm_addr: int = comm_addr
        self._running_loop: asyncio.AbstractEventLoop | None = None
        self._lock: asyncio.Lock | None = None
        self._timer: asyncio.TimerHandle | None = None
        self.timeout: int = timeout
        self.retries: int = retries
        self.keep_alive: bool = False
        self.protocol: asyncio.Protocol | None = None
        self.response_future: Future | None = None
        self.command: ProtocolCommand | None = None
        self._partial_data: bytes | None = None
        self._partial_missing: int = 0

    def _ensure_lock(self) -> asyncio.Lock:
        """Validate (or create) asyncio Lock.

           The asyncio.Lock must always be created from within's asyncio loop,
           so it cannot be eagerly created in constructor.
           Additionally, since asyncio.run() creates and closes its own loop,
           the lock's scope (its creating loop) mus be verified to support proper
           behavior in subsequent asyncio.run() invocations.
        """
        if self._lock and self._running_loop == asyncio.get_event_loop():
            return self._lock
        else:
            logger.debug("Creating lock instance for current event loop.")
            self._lock = asyncio.Lock()
            self._running_loop = asyncio.get_event_loop()
            self._close_transport()
            return self._lock

    def _max_retries_reached(self) -> Future:
        logger.debug("Max number of retries (%d) reached, request %s failed.", self.retries, self.command)
        self._close_transport()
        self.response_future = asyncio.get_running_loop().create_future()
        self.response_future.set_exception(MaxRetriesException)
        return self.response_future

    def _close_transport(self) -> None:
        if self._transport:
            try:
                self._transport.close()
            except RuntimeError:
                logger.debug("Failed to close transport.")
            self._transport = None
        # Cancel Future on connection lost
        if self.response_future and not self.response_future.done():
            self.response_future.cancel()

    async def close(self) -> None:
        """Close the underlying transport/connection."""
        raise NotImplementedError()

    async def send_request(self, command: ProtocolCommand) -> Future:
        """Convert command to request and send it to inverter."""
        raise NotImplementedError()

    def read_command(self, offset: int, count: int) -> ProtocolCommand:
        """Create read protocol command."""
        raise NotImplementedError()

    def write_command(self, register: int, value: int) -> ProtocolCommand:
        """Create write protocol command."""
        raise NotImplementedError()

    def write_multi_command(self, offset: int, values: bytes) -> ProtocolCommand:
        """Create write multiple protocol command."""
        raise NotImplementedError()


class UdpInverterProtocol(InverterProtocol, asyncio.DatagramProtocol):
    def __init__(self, host: str, port: int, comm_addr: int, timeout: int = 1, retries: int = 3):
        super().__init__(host, port, comm_addr, timeout, retries)
        self._transport: asyncio.transports.DatagramTransport | None = None
        self._retry: int = 0

    def read_command(self, offset: int, count: int) -> ProtocolCommand:
        """Create read protocol command."""
        return ModbusRtuReadCommand(self._comm_addr, offset, count)

    def write_command(self, register: int, value: int) -> ProtocolCommand:
        """Create write protocol command."""
        return ModbusRtuWriteCommand(self._comm_addr, register, value)

    def write_multi_command(self, offset: int, values: bytes) -> ProtocolCommand:
        """Create write multiple protocol command."""
        return ModbusRtuWriteMultiCommand(self._comm_addr, offset, values)

    async def _connect(self) -> None:
        if not self._transport or self._transport.is_closing():
            self._transport, self.protocol = await asyncio.get_running_loop().create_datagram_endpoint(
                lambda: self,
                remote_addr=(self._host, self._port),
            )

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        """On connection made"""
        self._transport = transport

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """On connection lost"""
        if exc:
            logger.debug("Socket closed with error: %s.", exc)
        else:
            logger.debug("Socket closed.")
        self._close_transport()

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        """On datagram received"""
        if self._timer:
            self._timer.cancel()
            self._timer = None
        try:
            if self._partial_data and self._partial_missing == len(data):
                logger.debug("Composed fragmented response: %s + %s", self._partial_data.hex(), data.hex())
                data = self._partial_data + data
                self._partial_data = None
                self._partial_missing = 0
            if self.command.validator(data):
                logger.debug("Received: %s", data.hex())
                self._retry = 0
                self.response_future.set_result(data)
            else:
                logger.debug("Received invalid response: %s", data.hex())
                asyncio.get_running_loop().call_soon(self._timeout_mechanism)
        except PartialResponseException as ex:
            logger.debug("Received response fragment (%d of %d): %s", ex.length, ex.expected, data.hex())
            self._partial_data = data
            self._partial_missing = ex.expected - ex.length
            self._timer = asyncio.get_running_loop().call_later(self.timeout, self._timeout_mechanism)
        except asyncio.InvalidStateError:
            logger.debug("Response already handled: %s", data.hex())
        except RequestRejectedException as ex:
            logger.debug("Received exception response: %s", data.hex())
            if self.response_future and not self.response_future.done():
                self.response_future.set_exception(ex)
            self._close_transport()

    def error_received(self, exc: Exception) -> None:
        """On error received"""
        logger.debug("Received error: %s", exc)
        self.response_future.set_exception(exc)
        self._close_transport()

    async def send_request(self, command: ProtocolCommand) -> Future:
        """Send message via transport"""
        await self._ensure_lock().acquire()
        try:
            await self._connect()
            response_future = asyncio.get_running_loop().create_future()
            self._send_request(command, response_future)
            await response_future
            return response_future
        except asyncio.CancelledError:
            if self._retry < self.retries:
                self._retry += 1
                if self._lock and self._lock.locked():
                    self._lock.release()
                if not self.keep_alive:
                    self._close_transport()
                return await self.send_request(command)
            else:
                return self._max_retries_reached()
        finally:
            if self._lock and self._lock.locked():
                self._lock.release()
            if not self.keep_alive:
                self._close_transport()

    def _send_request(self, command: ProtocolCommand, response_future: Future) -> None:
        """Send message via transport"""
        self.command = command
        self.response_future = response_future
        self._partial_data = None
        self._partial_missing = 0
        payload = command.request_bytes()
        if self._retry > 0:
            logger.debug("Sending: %s - retry #%s/%s", self.command, self._retry, self.retries)
        else:
            logger.debug("Sending: %s", self.command)
        self._transport.sendto(payload)
        self._timer = asyncio.get_running_loop().call_later(self.timeout, self._timeout_mechanism)

    def _timeout_mechanism(self) -> None:
        """Timeout mechanism to prevent hanging transport"""
        if self.response_future and self.response_future.done():
            logger.debug("Response already received.")
            self._retry = 0
        else:
            if self._timer:
                logger.debug("Failed to receive response to %s in time (%ds).", self.command, self.timeout)
                self._timer = None
            if self.response_future and not self.response_future.done():
                self.response_future.cancel()

    async def close(self):
        self._close_transport()


class TcpInverterProtocol(InverterProtocol, asyncio.Protocol):
    def __init__(self, host: str, port: int, comm_addr: int, timeout: int = 1, retries: int = 0):
        super().__init__(host, port, comm_addr, timeout, retries)
        self._transport: asyncio.transports.Transport | None = None
        self._retry: int = 0

    def read_command(self, offset: int, count: int) -> ProtocolCommand:
        """Create read protocol command."""
        return ModbusTcpReadCommand(self._comm_addr, offset, count)

    def write_command(self, register: int, value: int) -> ProtocolCommand:
        """Create write protocol command."""
        return ModbusTcpWriteCommand(self._comm_addr, register, value)

    def write_multi_command(self, offset: int, values: bytes) -> ProtocolCommand:
        """Create write multiple protocol command."""
        return ModbusTcpWriteMultiCommand(self._comm_addr, offset, values)

    async def _connect(self) -> None:
        if not self._transport or self._transport.is_closing():
            logger.debug("Opening connection.")
            self._transport, self.protocol = await asyncio.get_running_loop().create_connection(
                lambda: self,
                host=self._host, port=self._port,
            )
            if self.keep_alive:
                try:
                    sock = self._transport.get_extra_info('socket')
                    if sock is not None:
                        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 10)
                        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 10)
                        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 3)
                        if platform.system() == 'Windows':
                            sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, 10000, 10000))
                except AttributeError as ex:
                    logger.debug("Failed to apply KEEPALIVE: %s", ex)

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        """On connection made"""
        logger.debug("Connection opened.")
        pass

    def eof_received(self) -> None:
        logger.debug("EOF received.")
        self._close_transport()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        """On connection lost"""
        if exc:
            logger.debug("Connection closed with error: %s.", exc)
        else:
            logger.debug("Connection closed.")
        self._close_transport()

    def data_received(self, data: bytes) -> None:
        """On data received"""
        if self._timer:
            self._timer.cancel()
        try:
            if self._partial_data and self._partial_missing == len(data):
                logger.debug("Composed fragmented response: %s + %s", self._partial_data.hex(), data.hex())
                data = self._partial_data + data
                self._partial_data = None
                self._partial_missing = 0
            if self.command.validator(data):
                logger.debug("Received: %s", data.hex())
                self._retry = 0
                self.response_future.set_result(data)
            else:
                logger.debug("Received invalid response: %s", data.hex())
                self.response_future.set_exception(RequestRejectedException())
                self._close_transport()
        except PartialResponseException as ex:
            logger.debug("Received response fragment (%d of %d): %s", ex.length, ex.expected, data.hex())
            self._partial_data = data
            self._partial_missing = ex.expected - ex.length
            self._timer = asyncio.get_running_loop().call_later(self.timeout, self._timeout_mechanism)
        except asyncio.InvalidStateError:
            logger.debug("Response already handled: %s", data.hex())
        except RequestRejectedException as ex:
            logger.debug("Received exception response: %s", data.hex())
            if self.response_future and not self.response_future.done():
                self.response_future.set_exception(ex)
            # self._close_transport()

    def error_received(self, exc: Exception) -> None:
        """On error received"""
        logger.debug("Received error: %s", exc)
        self.response_future.set_exception(exc)
        self._close_transport()

    async def send_request(self, command: ProtocolCommand) -> Future:
        """Send message via transport"""
        await self._ensure_lock().acquire()
        try:
            await asyncio.wait_for(self._connect(), timeout=5)
            response_future = asyncio.get_running_loop().create_future()
            self._send_request(command, response_future)
            await response_future
            return response_future
        except asyncio.CancelledError:
            if self._retry < self.retries:
                if self._timer:
                    logger.debug("Connection broken error.")
                self._retry += 1
                if self._lock and self._lock.locked():
                    self._lock.release()
                self._close_transport()
                return await self.send_request(command)
            else:
                return self._max_retries_reached()
        except (ConnectionRefusedError, TimeoutError, OSError, asyncio.TimeoutError):
            if self._retry < self.retries:
                logger.debug("Connection refused error.")
                self._retry += 1
                if self._lock and self._lock.locked():
                    self._lock.release()
                return await self.send_request(command)
            else:
                return self._max_retries_reached()
        finally:
            if self._lock and self._lock.locked():
                self._lock.release()

    def _send_request(self, command: ProtocolCommand, response_future: Future) -> None:
        """Send message via transport"""
        self.command = command
        self.response_future = response_future
        self._partial_data = None
        self._partial_missing = 0
        payload = command.request_bytes()
        if self._retry > 0:
            logger.debug("Sending: %s - retry #%s/%s", self.command, self._retry, self.retries)
        else:
            logger.debug("Sending: %s", self.command)
        self._transport.write(payload)
        self._timer = asyncio.get_running_loop().call_later(self.timeout, self._timeout_mechanism)

    def _timeout_mechanism(self) -> None:
        """Retry mechanism to prevent hanging transport"""
        if self.response_future.done():
            self._retry = 0
        else:
            if self._timer:
                logger.debug("Failed to receive response to %s in time (%ds).", self.command, self.timeout)
                self._timer = None
            self._close_transport()

    async def close(self):
        await self._ensure_lock().acquire()
        try:
            self._close_transport()
        finally:
            if self._lock and self._lock.locked():
                self._lock.release()


class ProtocolResponse:
    """Definition of response to protocol command"""

    def __init__(self, raw_data: bytes, command: Optional[ProtocolCommand]):
        self.raw_data: bytes = raw_data
        self.command: ProtocolCommand = command
        self._bytes: io.BytesIO = io.BytesIO(self.response_data())

    def __repr__(self):
        return self.raw_data.hex()

    def response_data(self) -> bytes:
        if self.command is not None:
            return self.command.trim_response(self.raw_data)
        else:
            return self.raw_data

    def seek(self, address: int) -> None:
        if self.command is not None:
            self._bytes.seek(self.command.get_offset(address))
        else:
            self._bytes.seek(address)

    def read(self, size: int) -> bytes:
        return self._bytes.read(size)


class ProtocolCommand:
    """Definition of inverter protocol command"""

    def __init__(self, request: bytes, validator: Callable[[bytes], bool]):
        self.request: bytes = request
        self.validator: Callable[[bytes], bool] = validator

    def __eq__(self, other):
        if not isinstance(other, ProtocolCommand):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.request == other.request

    def __hash__(self):
        return hash(self.request)

    def __repr__(self):
        return self.request.hex()

    def request_bytes(self) -> bytes:
        """Return raw bytes payload, optionally pre-processed"""
        return self.request

    def trim_response(self, raw_response: bytes):
        """Trim raw response from header and checksum data"""
        return raw_response

    def get_offset(self, address: int):
        """Calculate relative offset to start of the response bytes"""
        return address

    async def execute(self, protocol: InverterProtocol) -> ProtocolResponse:
        """
        Execute the protocol command on the specified connection.

        Return ProtocolResponse with raw response data
        """
        try:
            response_future = await protocol.send_request(self)
            result = response_future.result()
            if result is not None:
                return ProtocolResponse(result, self)
            else:
                raise RequestFailedException(
                    "No response received to '" + self.request.hex() + "' request."
                )
        except (asyncio.CancelledError, ConnectionRefusedError):
            raise RequestFailedException(
                "No valid response received to '" + self.request.hex() + "' request."
            ) from None
        finally:
            if not protocol.keep_alive:
                await protocol.close()


class Aa55ProtocolCommand(ProtocolCommand):
    """
    Inverter communication protocol seen mostly on older generations of inverters.
    Quite probably it is some variation of the protocol used on RS-485 serial link,
    extended/adapted to UDP transport layer.

    Each request starts with header of 0xAA, 0x55, then 0xC0, 0x7F (client addr, inverter addr)
    followed by actual payload data.
    It is suffixed with 2 bytes of plain checksum of header+payload.

    Response starts again with 0xAA, 0x55, then 0x7F, 0xC0.
    5-6th bytes are some response type, byte 7 is length of the response payload.
    The last 2 bytes are again plain checksum of header+payload.
    """

    def __init__(self, payload: str, response_type: str, offset: int = 0, value: int = 0):
        super().__init__(
            bytes.fromhex(
                "AA55C07F"
                + payload
                + self._checksum(bytes.fromhex("AA55C07F" + payload)).hex()
            ),
            lambda x: self._validate_aa55_response(x, response_type),
        )
        self.first_address: int = offset
        self.value = value

    @staticmethod
    def _checksum(data: bytes) -> bytes:
        checksum = 0
        for each in data:
            checksum += each
        return checksum.to_bytes(2, byteorder="big", signed=False)

    @staticmethod
    def _validate_aa55_response(data: bytes, response_type: str) -> bool:
        """
        Validate the response.
        data[0:3] is header
        data[4:5] is response type
        data[6] is response payload length
        data[-2:] is checksum (plain sum of response data incl. header)
        """
        if len(data) <= 8:
            logger.debug("Response too short.")
            return False
        elif len(data) < data[6] + 9:
            raise PartialResponseException(len(data), data[6] + 9)
        elif len(data) > data[6] + 9:
            logger.debug("Response invalid - too long (%d).", len(data))
            return False
        elif response_type:
            data_rt_int = int.from_bytes(data[4:6], byteorder="big", signed=True)
            if int(response_type, 16) != data_rt_int:
                logger.debug("Response type unexpected: %04x, expected %s.", data_rt_int, response_type)
                return False

        checksum = 0
        for each in data[:-2]:
            checksum += each
        if checksum != int.from_bytes(data[-2:], byteorder="big", signed=True):
            logger.debug("Response checksum does not match.")
            return False
        return True

    def trim_response(self, raw_response: bytes):
        """Trim raw response from header and checksum data"""
        return raw_response[7:-2]

    def __repr__(self):
        if self.request[4] == 1:
            if self.request[5] == 2:
                return f'READ device info ({self.request.hex()})'
            elif self.request[5] == 6:
                return f'READ runtime data ({self.request.hex()})'
            elif self.request[5] == 9:
                return f'READ settings ({self.request.hex()})'
        else:
            return self.request.hex()


class Aa55ReadCommand(Aa55ProtocolCommand):
    """
    Inverter modbus READ command for retrieving <count> modbus registers starting at register # <offset>
    """

    def __init__(self, offset: int, count: int):
        super().__init__("011A03" + "{:04x}".format(offset) + "{:02x}".format(count), "019A", offset, count)

    def __repr__(self):
        if self.value > 1:
            return f'READ {self.value} registers from {self.first_address} ({self.request.hex()})'
        else:
            return f'READ register {self.first_address} ({self.request.hex()})'


class Aa55WriteCommand(Aa55ProtocolCommand):
    """
    Inverter aa55 WRITE command setting single register # <register> value <value>
    """

    def __init__(self, register: int, value: int):
        super().__init__("023905" + "{:04x}".format(register) + "01" + "{:04x}".format(value), "02B9", register, value)

    def __repr__(self):
        return f'WRITE {self.value} to register {self.first_address} ({self.request.hex()})'


class Aa55WriteMultiCommand(Aa55ProtocolCommand):
    """
    Inverter aa55 WRITE command setting multiple register # <register> value <value>
    """

    def __init__(self, offset: int, values: bytes):
        super().__init__("02390B" + "{:04x}".format(offset) + "{:02x}".format(len(values)) + values.hex(),
                         "02B9", offset, len(values) // 2)


class ModbusRtuProtocolCommand(ProtocolCommand):
    """
    Inverter communication protocol seen on newer generation of inverters, based on Modbus
    protocol over UDP transport layer.
    The modbus communication is rather simple, there are "registers" at specified addresses/offsets,
    each represented by 2 bytes. The protocol may query/update individual or range of these registers.
    Each register represents some measured value or operational settings.
    It's inverter implementation specific which register means what.
    Some values may span more registers (i.e. 4bytes measurement value over 2 registers).

    Every request usually starts with communication address (usually 0xF7, but can be changed).
    Second byte is the modbus command - 0x03 read multiple, 0x06 write single, 0x10 write multiple.
    Bytes 3-4 represent the register address (or start of range)
    Bytes 5-6 represent the command parameter (range size or actual value for write).
    Last 2 bytes of request is the CRC-16 (modbus flavor) of the request.

    Responses seem to always start with 0xAA, 0x55, then the comm_addr and modbus command.
    (If the command fails, the highest bit of command is set to 1 ?)
    For read requests, next byte is response payload length, then the actual payload.
    Last 2 bytes of response is again the CRC-16 of the response.
    """

    def __init__(self, request: bytes, cmd: int, offset: int, value: int):
        super().__init__(
            request,
            lambda x: validate_modbus_rtu_response(x, cmd, offset, value),
        )
        self.first_address: int = offset
        self.value = value

    def trim_response(self, raw_response: bytes):
        """Trim raw response from header and checksum data"""
        return raw_response[5:-2]

    def get_offset(self, address: int):
        """Calculate relative offset to start of the response bytes"""
        return (address - self.first_address) * 2


class ModbusRtuReadCommand(ModbusRtuProtocolCommand):
    """
    Inverter Modbus/RTU READ command for retrieving <count> modbus registers starting at register # <offset>
    """

    def __init__(self, comm_addr: int, offset: int, count: int):
        super().__init__(
            create_modbus_rtu_request(comm_addr, MODBUS_READ_CMD, offset, count),
            MODBUS_READ_CMD, offset, count)

    def __repr__(self):
        if self.value > 1:
            return f'READ {self.value} registers from {self.first_address} ({self.request.hex()})'
        else:
            return f'READ register {self.first_address} ({self.request.hex()})'


class ModbusRtuWriteCommand(ModbusRtuProtocolCommand):
    """
    Inverter Modbus/RTU WRITE command setting single modbus register # <register> value <value>
    """

    def __init__(self, comm_addr: int, register: int, value: int):
        super().__init__(
            create_modbus_rtu_request(comm_addr, MODBUS_WRITE_CMD, register, value),
            MODBUS_WRITE_CMD, register, value)

    def __repr__(self):
        return f'WRITE {self.value} to register {self.first_address} ({self.request.hex()})'


class ModbusRtuWriteMultiCommand(ModbusRtuProtocolCommand):
    """
    Inverter Modbus/RTU WRITE command setting multiple modbus register # <register> value <value>
    """

    def __init__(self, comm_addr: int, offset: int, values: bytes):
        super().__init__(
            create_modbus_rtu_multi_request(comm_addr, MODBUS_WRITE_MULTI_CMD, offset, values),
            MODBUS_WRITE_MULTI_CMD, offset, len(values) // 2)


class ModbusTcpProtocolCommand(ProtocolCommand):
    """
    Modbus/TCP inverter communication protocol.
    """

    def __init__(self, request: bytes, cmd: int, offset: int, value: int):
        super().__init__(
            request,
            lambda x: validate_modbus_tcp_response(x, cmd, offset, value),
        )
        self.first_address: int = offset
        self.value = value

    def request_bytes(self) -> bytes:
        """Return raw bytes payload, optionally pre-processed"""
        # Apply sequential Modbus/TCP transaction identifier
        self.request = _next_tx() + self.request[2:]
        return self.request

    def trim_response(self, raw_response: bytes):
        """Trim raw response from header and checksum data"""
        return raw_response[9:]

    def get_offset(self, address: int):
        """Calculate relative offset to start of the response bytes"""
        return (address - self.first_address) * 2


class ModbusTcpReadCommand(ModbusTcpProtocolCommand):
    """
    Inverter Modbus/TCP READ command for retrieving <count> modbus registers starting at register # <offset>
    """

    def __init__(self, comm_addr: int, offset: int, count: int):
        super().__init__(
            create_modbus_tcp_request(comm_addr, MODBUS_READ_CMD, offset, count),
            MODBUS_READ_CMD, offset, count)

    def __repr__(self):
        if self.value > 1:
            return f'READ {self.value} registers from {self.first_address} ({self.request.hex()})'
        else:
            return f'READ register {self.first_address} ({self.request.hex()})'


class ModbusTcpWriteCommand(ModbusTcpProtocolCommand):
    """
    Inverter Modbus/TCP WRITE command setting single modbus register # <register> value <value>
    """

    def __init__(self, comm_addr: int, register: int, value: int):
        super().__init__(
            create_modbus_tcp_request(comm_addr, MODBUS_WRITE_CMD, register, value),
            MODBUS_WRITE_CMD, register, value)

    def __repr__(self):
        return f'WRITE {self.value} to register {self.first_address} ({self.request.hex()})'


class ModbusTcpWriteMultiCommand(ModbusTcpProtocolCommand):
    """
    Inverter Modbus/TCP WRITE command setting multiple modbus register # <register> value <value>
    """

    def __init__(self, comm_addr: int, offset: int, values: bytes):
        super().__init__(
            create_modbus_tcp_multi_request(comm_addr, MODBUS_WRITE_MULTI_CMD, offset, values),
            MODBUS_WRITE_MULTI_CMD, offset, len(values) // 2)



import asyncio


class EchoServerProtocol(asyncio.Protocol):
    def connection_made(self, transport):
        peername = transport.get_extra_info('peername')
        print('Connection from {}'.format(peername))
        self.transport = transport

    def datagram_received(self, data, addr):
        payload = data.hex()
        print(payload)

        if payload == "aa55c07f0102000241":
            # Device info
            self.transport.sendto(bytes.fromhex(
                "aa557fc001824d323532354b4757353034382d4553412331300000000000000000000000000039353034384553413030305730303030333630303431302d30343032352d3235203431302d30323033342d323001102f"),
                addr)
        elif payload == "aa55c07f0106000245":
            # Running data
            self.transport.sendto(bytes.fromhex(
                "aa557fc001868c09270047020fe60042010214000200500118000400000032000064006464020000010a11009e0ce11389010a11000303e11389010202010000000000023780000053c3012600770001ad1510c30100200100000001000003e500000840000018051a0e0e120000000000000000000000000000000000000000000000000000ca260000baab0200000000000012e1"),
                addr)
        elif payload == "aa55c07f0109000248":
            # Settings data
            self.transport.sendto(bytes.fromhex(
                "aa557fc00189560000000000000000000000000001000100010000000a00d2024000f0008001e0000a000f0000000000000064024003e8001e00f20000000000000000023f000000070000000000000001038403e801e0000a000000220c12"),
                addr)
        elif payload == "aa55c07f011a030701040268":
            # Read eco_mode_1
            self.transport.sendto(bytes.fromhex(
                "aa557fc0019a08000000000000007f0360"),
                addr)
        elif payload == "aa55c07f032c0500000000000272":
            self.transport.sendto(bytes.fromhex(
                "aa557fc003ac010602f4"),
                addr)
        elif payload == "aa55c07f032d0500000000000273":
            self.transport.sendto(bytes.fromhex(
                "aa557fc003ad010602f5"),
                addr)
        elif payload == "aa55c07f02390507000100010287":
            self.transport.sendto(bytes.fromhex(
                "aa557fc002b901060300"),
                addr)
        elif payload == "aa55c07f033601000278":
            self.transport.sendto(bytes.fromhex(
                "aa557fc003b6010602fe"),
                addr)
        elif payload == "aa55c07f03590100029b":
            self.transport.sendto(bytes.fromhex(
                "aa557fc003d901060321"),
                addr)
        else:
            self.transport.sendto(bytes.fromhex("00000000"), addr)
        # print('Close the client socket')
        # self.transport.close()


async def main():
    # Get a reference to the event loop as we plan to use
    # low-level APIs.
    loop = asyncio.get_running_loop()

    transport, protocol = await loop.create_datagram_endpoint(
        lambda: EchoServerProtocol(),
        local_addr=('127.0.0.1', 8899))

    try:
        await asyncio.sleep(3600)  # Serve for 1 hour.
    finally:
        transport.close()


asyncio.run(main())
